# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import ssl
import urllib.error

import pytest

from awscli.customizations.exceptions import ConfigurationError
from awscli.customizations.sso.resolve import (
    _extract_region_from_hostname,
    _follow_redirect,
    _normalize_url,
    is_aws_owned_domain,
    resolve_start_url,
)
from awscli.testutils import mock


class TestIsAwsOwnedDomain:
    @pytest.mark.parametrize(
        'hostname',
        [
            'ssoins-abc123.portal.us-west-2.app.aws',
            'ssoins-abc123.portal.us-east-1.app.aws',
            'SSOINS-ABC123.PORTAL.US-WEST-2.APP.AWS',
            'ssoins-abc123.us-west-2.portal.amazonaws.com',
            'd-abc123.awsapps.com',
            'myalias.awsapps.com',
            'awsapps.com',
            'identitycenter.amazonaws.com',
            'identitycenter.amazonaws.com.',
            'd-abc123.awsapps.cn',
            'myalias.awsapps.cn',
            'ssoins-abc123.cn-north-1.portal.amazonaws.com.cn',
            'ssoins-abc123.cn-northwest-1.portal.amazonaws.com.cn',
            'identitycenter.amazonaws.com.cn',
            'ssoins-abc123.portal.cn-north-1.app.amazonwebservices.com.cn',
            'ssoins-abc123.eusc-de-east-1.portal.amazonaws.eu',
            'ssoins-abc123.portal.eusc-de-east-1.api.amazonwebservices.eu',
            'identitycenter.amazonaws.eu',
            'identitycenter.us-gov.amazonaws.com',
        ],
    )
    def test_aws_owned_returns_true(self, hostname):
        assert is_aws_owned_domain(hostname) is True

    @pytest.mark.parametrize(
        'hostname',
        [
            'aws.mycompany.com',
            'awsapps.com.evil.net',
            'evil-awsapps.com',
            'x.app.aws.example.com',
            'amazonaws.com.example.net',
            'portal.amazonaws.com.evil.net',
            'notidentitycenter.amazonaws.com',
            'example.com',
            '',
            'awsapps.cn.evil.net',
            'evil-awsapps.cn',
            'portal.amazonaws.com.cn.evil.net',
        ],
    )
    def test_non_aws_owned_returns_false(self, hostname):
        assert is_aws_owned_domain(hostname) is False


class TestExtractRegionFromHostname:
    @pytest.mark.parametrize(
        'hostname, expected_region',
        [
            ('ssoins-abc.portal.us-west-2.app.aws', 'us-west-2'),
            ('ssoins-abc.portal.us-east-1.app.aws', 'us-east-1'),
            ('ssoins-abc.portal.eu-west-1.app.aws', 'eu-west-1'),
            ('ssoins-abc.portal.cn-north-1.app.aws', 'cn-north-1'),
            ('ssoins-abc.us-gov-west-1.portal.amazonaws.com', 'us-gov-west-1'),
            ('ssoins-abc.us-west-2.portal.amazonaws.com', 'us-west-2'),
            ('ssoins-abc.cn-north-1.portal.amazonaws.com.cn', 'cn-north-1'),
            (
                'ssoins-abc.cn-northwest-1.portal.amazonaws.com.cn',
                'cn-northwest-1',
            ),
            (
                'ssoins-abc.portal.cn-north-1.app.amazonwebservices.com.cn',
                'cn-north-1',
            ),
            (
                'ssoins-abc.eusc-de-east-1.portal.amazonaws.eu',
                'eusc-de-east-1',
            ),
            (
                'ssoins-abc.portal.eusc-de-east-1.api.amazonwebservices.eu',
                'eusc-de-east-1',
            ),
        ],
    )
    def test_extracts_region(self, hostname, expected_region):
        assert _extract_region_from_hostname(hostname) == expected_region

    @pytest.mark.parametrize(
        'hostname',
        [
            'd-abc123.awsapps.com',
            'd-abc123.awsapps.cn',
            'identitycenter.amazonaws.com',
            'identitycenter.amazonaws.com.cn',
            'identitycenter.amazonaws.eu',
            'identitycenter.us-gov.amazonaws.com',
            'aws.mycompany.com',
        ],
    )
    def test_returns_none_for_region_less_hostnames(self, hostname):
        assert _extract_region_from_hostname(hostname) is None


class TestNormalizeUrl:
    @pytest.mark.parametrize(
        'url, expected',
        [
            ('https://example.com:443/', 'https://example.com'),
            ('https://example.com:443/path/', 'https://example.com/path'),
            ('https://example.com/', 'https://example.com'),
            ('https://example.com:8443/path', 'https://example.com:8443/path'),
            ('http://example.com:80/', 'http://example.com'),
            ('https://example.com', 'https://example.com'),
        ],
    )
    def test_normalize_url(self, url, expected):
        assert _normalize_url(url) == expected


class TestResolveStartUrl:
    def _mock_session(self):
        return mock.Mock()

    def test_aws_owned_url_returns_configured_region(self):
        session = self._mock_session()
        resolved_url, region = resolve_start_url(
            'https://ssoins-abc.portal.us-west-2.app.aws',
            session=session,
            configured_region='us-west-2',
        )
        assert resolved_url == 'https://ssoins-abc.portal.us-west-2.app.aws'
        assert region == 'us-west-2'

    def test_aws_owned_url_no_network_call(self):
        session = self._mock_session()
        with mock.patch('urllib.request.build_opener') as mock_opener:
            resolve_start_url(
                'https://ssoins-abc.portal.us-west-2.app.aws',
                session=session,
                configured_region='us-west-2',
            )
            mock_opener.assert_not_called()

    def test_aws_owned_url_without_configured_region_raises_error(self):
        session = self._mock_session()
        with pytest.raises(
            ConfigurationError,
            match='Missing required configuration: sso_region',
        ):
            resolve_start_url(
                'https://ssoins-abc.portal.us-west-2.app.aws',
                session=session,
            )

    def test_aws_owned_url_uses_configured_region_not_hostname(self):
        session = self._mock_session()
        resolved_url, region = resolve_start_url(
            'https://ssoins-abc.portal.us-west-2.app.aws',
            session=session,
            configured_region='us-east-1',
        )
        assert resolved_url == 'https://ssoins-abc.portal.us-west-2.app.aws'
        assert region == 'us-east-1'

    def test_awsapps_url_requires_configured_region(self):
        session = self._mock_session()
        with pytest.raises(
            ConfigurationError,
            match='Missing required configuration: sso_region',
        ):
            resolve_start_url(
                'https://d-abc123.awsapps.com/start',
                session=session,
            )

    def test_awsapps_url_uses_configured_region(self):
        session = self._mock_session()
        resolved_url, region = resolve_start_url(
            'https://d-abc123.awsapps.com/start',
            session=session,
            configured_region='us-east-1',
        )
        assert resolved_url == 'https://d-abc123.awsapps.com/start'
        assert region == 'us-east-1'

    def test_http_scheme_raises_error(self):
        session = self._mock_session()
        with pytest.raises(ConfigurationError, match='https scheme'):
            resolve_start_url('http://aws.mycompany.com', session=session)

    def test_invalid_url_raises_error(self):
        session = self._mock_session()
        with pytest.raises(ConfigurationError, match='Invalid sso_start_url'):
            resolve_start_url('https://', session=session)

    def test_vanity_url_follows_redirect(self):
        session = self._mock_session()
        redirect_url = 'https://ssoins-abc.portal.us-east-1.app.aws:443/'
        normalized_url = 'https://ssoins-abc.portal.us-east-1.app.aws'

        with mock.patch(
            'awscli.customizations.sso.resolve._follow_redirect'
        ) as mock_follow:
            mock_follow.return_value = redirect_url
            resolved_url, region = resolve_start_url(
                'https://aws.mycompany.com',
                session=session,
            )
            assert resolved_url == normalized_url
            assert region == 'us-east-1'
            mock_follow.assert_called_once_with(
                'https://aws.mycompany.com',
                timeout=10,
                verify=None,
            )

    def test_vanity_url_uses_parsed_region_not_configured(self):
        session = self._mock_session()
        with mock.patch(
            'awscli.customizations.sso.resolve._follow_redirect'
        ) as mock_follow:
            mock_follow.return_value = (
                'https://ssoins-abc.portal.us-west-2.app.aws'
            )
            resolved_url, region = resolve_start_url(
                'https://aws.mycompany.com',
                session=session,
                configured_region='eu-west-1',
            )
            assert region == 'us-west-2'

    def test_vanity_url_resolves_to_non_aws_domain_raises_error(self):
        session = self._mock_session()
        with mock.patch(
            'awscli.customizations.sso.resolve._follow_redirect'
        ) as mock_follow:
            mock_follow.return_value = 'https://not-aws.example.com'
            with pytest.raises(ConfigurationError, match='Could not resolve'):
                resolve_start_url('https://aws.mycompany.com', session=session)

    def test_vanity_url_resolves_to_http_raises_error(self):
        session = self._mock_session()
        with mock.patch(
            'awscli.customizations.sso.resolve._follow_redirect'
        ) as mock_follow:
            mock_follow.return_value = (
                'http://ssoins-abc.portal.us-east-1.app.aws'
            )
            with pytest.raises(ConfigurationError, match='must use https'):
                resolve_start_url('https://aws.mycompany.com', session=session)

    def test_vanity_url_region_less_requires_configured_region(self):
        session = self._mock_session()
        with mock.patch(
            'awscli.customizations.sso.resolve._follow_redirect'
        ) as mock_follow:
            mock_follow.return_value = 'https://d-abc123.awsapps.com/start'
            with pytest.raises(
                ConfigurationError, match='Cannot determine region'
            ):
                resolve_start_url('https://aws.mycompany.com', session=session)

    def test_vanity_url_region_less_uses_configured_region(self):
        session = self._mock_session()
        with mock.patch(
            'awscli.customizations.sso.resolve._follow_redirect'
        ) as mock_follow:
            mock_follow.return_value = 'https://d-abc123.awsapps.com/start'
            resolved_url, region = resolve_start_url(
                'https://aws.mycompany.com',
                session=session,
                configured_region='us-east-1',
            )
            assert region == 'us-east-1'

    def test_custom_timeout_passed_to_follow_redirect(self):
        session = self._mock_session()
        with mock.patch(
            'awscli.customizations.sso.resolve._follow_redirect'
        ) as mock_follow:
            mock_follow.return_value = (
                'https://ssoins-abc.portal.us-east-1.app.aws'
            )
            resolve_start_url(
                'https://aws.mycompany.com',
                session=session,
                timeout=5,
            )
            mock_follow.assert_called_once_with(
                'https://aws.mycompany.com',
                timeout=5,
                verify=None,
            )


class TestFollowRedirect:
    def _make_http_error(self, code, headers=None):
        if headers is None:
            headers = {}
        import http.client

        msg = http.client.HTTPMessage()
        for k, v in headers.items():
            msg[k] = v
        return urllib.error.HTTPError(
            url='https://example.com',
            code=code,
            msg='',
            hdrs=msg,
            fp=None,
        )

    def test_head_redirect_returns_location(self):
        redirect_target = 'https://ssoins-abc.portal.us-east-1.app.aws'
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_opener.open.side_effect = self._make_http_error(
                302, {'Location': redirect_target}
            )
            result = _follow_redirect('https://aws.mycompany.com')
            assert result == redirect_target

    @pytest.mark.parametrize('head_error_code', [405, 501])
    def test_head_unsupported_falls_back_to_get(self, head_error_code):
        redirect_target = 'https://ssoins-abc.portal.us-east-1.app.aws'
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_opener.open.side_effect = [
                self._make_http_error(head_error_code),
                self._make_http_error(302, {'Location': redirect_target}),
            ]
            result = _follow_redirect('https://aws.mycompany.com')
            assert result == redirect_target
            assert mock_opener.open.call_count == 2

    def test_head_200_returns_original_url(self):
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_response = mock.Mock()
            mock_opener.open.return_value = mock_response
            result = _follow_redirect('https://aws.mycompany.com')
            assert result == 'https://aws.mycompany.com'
            mock_response.close.assert_called_once()

    def test_missing_location_header_raises_error(self):
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_opener.open.side_effect = self._make_http_error(302, {})
            with pytest.raises(ConfigurationError, match='missing Location'):
                _follow_redirect('https://aws.mycompany.com')

    def test_non_redirect_error_raises_configuration_error(self):
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_opener.open.side_effect = self._make_http_error(500)
            with pytest.raises(ConfigurationError, match='HTTP 500'):
                _follow_redirect('https://aws.mycompany.com')

    def test_url_error_raises_configuration_error(self):
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_opener.open.side_effect = urllib.error.URLError(
                'Name or service not known'
            )
            with pytest.raises(ConfigurationError, match='Name or service'):
                _follow_redirect('https://aws.mycompany.com')

    def test_relative_location_is_resolved(self):
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_opener.open.side_effect = self._make_http_error(
                302, {'Location': '/start'}
            )
            result = _follow_redirect('https://aws.mycompany.com/portal')
            assert result == 'https://aws.mycompany.com/start'

    def test_stops_after_max_redirects(self):
        intermediate = 'https://intermediate.example.com/path'
        with mock.patch('urllib.request.build_opener') as mock_build:
            mock_opener = mock.Mock()
            mock_build.return_value = mock_opener
            mock_opener.open.side_effect = self._make_http_error(
                302, {'Location': intermediate}
            )
            result = _follow_redirect('https://aws.mycompany.com')
            assert result == intermediate
            assert mock_opener.open.call_count == 1

    def test_verify_false_disables_ssl(self):
        with mock.patch('ssl.create_default_context') as mock_ctx_factory:
            mock_ctx = mock.Mock()
            mock_ctx_factory.return_value = mock_ctx
            with mock.patch('urllib.request.build_opener'):
                with mock.patch('urllib.request.HTTPSHandler'):
                    try:
                        _follow_redirect(
                            'https://aws.mycompany.com', verify=False
                        )
                    except Exception:
                        pass
            assert mock_ctx.check_hostname is False
            assert mock_ctx.verify_mode == ssl.CERT_NONE

    def test_verify_ca_bundle_path_loaded(self, tmp_path):
        ca_file = tmp_path / 'ca.pem'
        ca_file.write_text('fake cert')
        with mock.patch('ssl.create_default_context') as mock_ctx_factory:
            mock_ctx = mock.Mock()
            mock_ctx_factory.return_value = mock_ctx
            with mock.patch('urllib.request.build_opener'):
                with mock.patch('urllib.request.HTTPSHandler'):
                    try:
                        _follow_redirect(
                            'https://aws.mycompany.com',
                            verify=str(ca_file),
                        )
                    except Exception:
                        pass
            mock_ctx.load_verify_locations.assert_called_once_with(
                cafile=str(ca_file)
            )

    def test_verify_none_uses_bundled_ca(self):
        with mock.patch('ssl.create_default_context') as mock_ctx_factory:
            mock_ctx = mock.Mock()
            mock_ctx_factory.return_value = mock_ctx
            with mock.patch('urllib.request.build_opener'):
                with mock.patch('urllib.request.HTTPSHandler'):
                    with mock.patch(
                        'awscli.customizations.sso.resolve.get_cert_path',
                        return_value='/bundled/cacert.pem',
                    ) as mock_get_cert_path:
                        try:
                            _follow_redirect('https://aws.mycompany.com')
                        except Exception:
                            pass
            mock_get_cert_path.assert_called_once_with(True)
            mock_ctx.load_verify_locations.assert_called_once_with(
                cafile='/bundled/cacert.pem'
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
from io import StringIO
from unittest.mock import Mock

import pytest

from awscli.customizations.agenttoolkit.agents import SKILL_METADATA_FILENAME
from awscli.customizations.agenttoolkit.remove_skill import RemoveCommand
from awscli.customizations.exceptions import ParamValidationError
from tests.unit.customizations.agenttoolkit.utils import (
    make_config,
    make_skill,
)


def _run_remove(agent_configs, args):
    stream = StringIO()
    session = Mock()
    session.user_agent_extra = ''
    session.emit_first_non_none_response.return_value = None
    cmd = RemoveCommand(session, stream=stream, agent_configs=agent_configs)
    rc = cmd(args=args, parsed_globals=Mock())
    return rc, stream.getvalue()


def test_remove_skill_not_installed(tmp_path):
    (tmp_path / '.test-agent' / 'skills').mkdir(parents=True)
    configs = [make_config(tmp_path)]
    with pytest.raises(ParamValidationError, match='not installed'):
        _run_remove(configs, ['--skill-name', 'aws-foo'])


def test_remove_skill_success(tmp_path):
    make_skill(tmp_path, '.test-agent', 'aws-serverless')
    configs = [make_config(tmp_path)]
    rc, output = _run_remove(configs, ['--skill-name', 'aws-serverless'])
    assert rc == 0
    assert 'Removed aws-serverless from Test Agent' in output
    assert not (
        tmp_path / '.test-agent' / 'skills' / 'aws-serverless'
    ).exists()


def test_remove_from_multiple_agents(tmp_path):
    make_skill(tmp_path, '.agent-a', 'aws-serverless')
    make_skill(tmp_path, '.agent-b', 'aws-serverless')
    configs = [
        make_config(
            tmp_path,
            id='agent-a',
            display_name='Agent A',
            detection_path=str(tmp_path / '.agent-a'),
        ),
        make_config(
            tmp_path,
            id='agent-b',
            display_name='Agent B',
            detection_path=str(tmp_path / '.agent-b'),
        ),
    ]
    rc, output = _run_remove(configs, ['--skill-name', 'aws-serverless'])
    assert rc == 0
    assert 'Agent A' in output
    assert 'Agent B' in output
    assert not (tmp_path / '.agent-a' / 'skills' / 'aws-serverless').exists()
    assert not (tmp_path / '.agent-b' / 'skills' / 'aws-serverless').exists()


def test_remove_with_client_filter(tmp_path):
    make_skill(tmp_path, '.agent-a', 'aws-serverless')
    make_skill(tmp_path, '.agent-b', 'aws-serverless')
    configs = [
        make_config(
            tmp_path,
            id='agent-a',
            display_name='Agent A',
            detection_path=str(tmp_path / '.agent-a'),
        ),
        make_config(
            tmp_path,
            id='agent-b',
            display_name='Agent B',
            detection_path=str(tmp_path / '.agent-b'),
        ),
    ]
    rc, output = _run_remove(
        configs,
        ['--skill-name', 'aws-serverless', '--agent', 'agent-a'],
    )
    assert rc == 0
    assert 'Agent A' in output
    assert 'Agent B' not in output
    assert not (tmp_path / '.agent-a' / 'skills' / 'aws-serverless').exists()
    assert (tmp_path / '.agent-b' / 'skills' / 'aws-serverless').exists()


def test_remove_skill_dedupes_shared_skills_path(tmp_path):
    (tmp_path / '.agent-a').mkdir()
    (tmp_path / '.agent-b').mkdir()
    shared = tmp_path / 'shared' / 'skills'
    skill_dir = shared / 'aws-s3'
    skill_dir.mkdir(parents=True)
    (skill_dir / 'SKILL.md').write_text('test')
    (skill_dir / SKILL_METADATA_FILENAME).write_text(
        json.dumps({'version': 'v1'})
    )
    configs = [
        make_config(
            tmp_path,
            id='a',
            display_name='Agent A',
            detection_path=str(tmp_path / '.agent-a'),
            skills_path_override=str(shared),
        ),
        make_config(
            tmp_path,
            id='b',
            display_name='Agent B',
            detection_path=str(tmp_path / '.agent-b'),
            skills_path_override=str(shared),
        ),
    ]
    rc, output = _run_remove(configs, ['--skill-name', 'aws-s3'])
    assert rc == 0
    assert output.count('Removed aws-s3') == 1
    assert not (shared / 'aws-s3').exists()


def test_remove_skill_universal_row_coexists_with_override_agent(tmp_path):
    (tmp_path / '.codex').mkdir()
    universal_base = tmp_path / '.agents'
    shared = universal_base / 'skills'
    skill_dir = shared / 'aws-cdk'
    skill_dir.mkdir(parents=True)
    (skill_dir / 'SKILL.md').write_text('test')
    (skill_dir / SKILL_METADATA_FILENAME).write_text(
        json.dumps({'version': 'v1'})
    )
    configs = [
        make_config(
            tmp_path,
            id='codex',
            display_name='Codex',
            detection_path=str(tmp_path / '.codex'),
            skills_path_override=str(shared),
        ),
        make_config(
            tmp_path,
            id='universal',
            display_name='Universal (Codex)',
            detection_path=str(universal_base),
        ),
    ]
    rc, output = _run_remove(configs, ['--skill-name', 'aws-cdk'])
    assert rc == 0
    assert 'Removed aws-cdk from Universal (Codex)' in output
    # Codex shares the skills path; the universal row owns the print line.
    assert output.count('Removed aws-cdk') == 1
    assert not (shared / 'aws-cdk').exists()


def test_remove_skill_without_marker_refused(tmp_path):
    make_skill(tmp_path, '.test-agent', 'hand-rolled', with_marker=False)
    configs = [make_config(tmp_path)]
    with pytest.raises(ParamValidationError, match='not installed'):
        _run_remove(configs, ['--skill-name', 'hand-rolled'])
    assert (tmp_path / '.test-agent' / 'skills' / 'hand-rolled').exists()


def test_remove_invalid_client_rejected(tmp_path):
    configs = [make_config(tmp_path)]
    with pytest.raises(ParamValidationError, match='Invalid agent'):
        _run_remove(
            configs,
            ['--skill-name', 'aws-serverless', '--agent', 'nonexistent'],
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.compat import compat_input
from awscli.customizations.exceptions import ConfigurationError
from awscli.customizations.utils import uni_print
from awscli.utils import is_stdin_a_tty


def yes_no_choice(prompt):
    """
    Prompts the user to answer a yes/no question.
    Continually re-prompts for invalid selections.

    :param prompt: Prompt text.
    :returns: True for yes, False for no.
    """
    while True:
        response = compat_input(prompt)

        if response.lower() in ('y', 'yes'):
            return True
        elif response.lower() in ('n', 'no'):
            return False
        else:
            uni_print('Invalid response. Please enter "y" or "n"\n')


def multiselect_choice(
    message,
    items,
    display_format=None,
    preselected=None,
    pt_input=None,
    pt_output=None,
):
    """
    Present a list of items with checkboxes for multi-selection.

    Arrow keys to navigate, space to toggle, enter to confirm.

    :param message: Prompt message displayed above the choices.
    :param items: List of items to select from.
    :param display_format: Optional callable to format items for display.
    :param preselected: Optional set of indices to pre-check.
    :param pt_input: Optional prompt_toolkit input (for testing).
    :param pt_output: Optional prompt_toolkit output (for testing).
    :returns: List of selected items.
    """
    if pt_input is None and not is_stdin_a_tty():
        raise ConfigurationError(
            "This command requires an interactive terminal (TTY)."
        )
    # Imported lazily so that loading this module does not pull in
    # prompt_toolkit for commands that only use yes_no_choice.
    from prompt_toolkit import Application
    from prompt_toolkit.key_binding.key_bindings import KeyBindings
    from prompt_toolkit.layout import HSplit, Layout, Window
    from prompt_toolkit.layout.controls import UIContent, UIControl
    from prompt_toolkit.layout.dimension import Dimension
    from prompt_toolkit.layout.screen import Point
    from prompt_toolkit.widgets import Label

    class MultiSelectControl(UIControl):
        def __init__(self, items, display_format=None, preselected=None):
            self._items = items
            self._cursor = 0
            self._selected = set(
                preselected if preselected is not None else range(len(items))
            )
            self._display_format = display_format

        def is_focusable(self):
            return True

        def preferred_width(self, max_width):
            return max_width

        def preferred_height(
            self, width, max_height, wrap_lines, get_line_prefix
        ):
            return len(self._items)

        def create_content(self, width, height):
            def get_line(i):
                check = 'x' if i in self._selected else ' '
                item = self._items[i]
                if self._display_format:
                    item = self._display_format(item)
                style = 'bold' if i == self._cursor else ''
                return [(style, f'  [{check}] {item}')]

            return UIContent(
                get_line=get_line,
                cursor_position=Point(x=0, y=self._cursor),
                line_count=len(self._items),
            )

        def get_key_bindings(self):
            kb = KeyBindings()

            @kb.add('up')
            def move_up(event):
                self._cursor = (self._cursor - 1) % len(self._items)

            @kb.add('down')
            def move_down(event):
                self._cursor = (self._cursor + 1) % len(self._items)

            @kb.add(' ')
            def toggle(event):
                if self._cursor in self._selected:
                    self._selected.discard(self._cursor)
                else:
                    self._selected.add(self._cursor)

            @kb.add('enter')
            def confirm(event):
                result = [self._items[i] for i in sorted(self._selected)]
                event.app.exit(result=result)

            return kb

    control = MultiSelectControl(
        items, display_format=display_format, preselected=preselected
    )
    body = Window(
        control,
        always_hide_cursor=False,
        height=Dimension(min=len(items), max=len(items)),
    )

    layout = Layout(
        HSplit(
            [
                Label(
                    f'{message} (space to toggle, up/down arrows to navigate, enter to confirm):'
                ),
                body,
            ]
        ),
        focused_element=body,
    )

    app_bindings = KeyBindings()

    @app_bindings.add('c-c')
    def exit_app(event):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    app = Application(
        layout=layout,
        key_bindings=app_bindings,
        full_screen=False,
        erase_when_done=True,
        input=pt_input,
        output=pt_output,
    )
    return app.run()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import dataclasses
import enum
import glob
import json
import os
import shutil
import subprocess
from typing import Optional

SKILL_FILENAME = 'SKILL.md'
SKILL_METADATA_FILENAME = '.aws-skill-metadata'
AWS_MCP_SERVER_KEY = 'aws-mcp'
UNIVERSAL_ROW_ID = 'universal'


def collapse_home(path):
    home = os.path.expanduser('~').rstrip(os.sep)
    if path == home:
        return '~'
    if path.startswith(home + os.sep):
        return '~' + path[len(home) :]
    return path


class McpConfigureAction(enum.Enum):
    CONFIGURED = 'configured'
    ALREADY_CONFIGURED = 'already_configured'
    SKIPPED = 'skipped'


_AWS_MCP_PROXY_ARGS = [
    'mcp-proxy-for-aws@latest',
    'https://aws-mcp.us-east-1.api.aws/mcp',
    '--metadata',
    'INSTALL_SOURCE=aws-cli',
]


DEFAULT_MCP_SERVER_CONFIG = {
    'command': 'uvx',
    'args': _AWS_MCP_PROXY_ARGS,
}


@dataclasses.dataclass
class AgentConfig:
    """Definition of a supported AI coding agent.

    To add a new agent, append an instance to ``AGENT_CONFIGS`` in this
    module. Most agents only need ``id``, ``display_name``, and
    ``detection_path``; the remaining fields cover variations across
    different agents' filesystem layouts and MCP config formats.
    """

    id: str
    """Stable identifier. Used as the value for ``--agent`` filters."""

    display_name: str
    """Human-readable name shown in wizard output and command results."""

    detection_path: str
    """Filesystem path probed to decide whether the agent is installed.
    Tilde expansion is supported. The agent is considered detected when
    this path resolves to an existing directory."""

    detection_path_env_override: Optional[str] = None
    """Environment variable name that overrides ``detection_path`` when
    set (e.g. ``CLAUDE_CONFIG_DIR``)."""

    skills_dir: str = 'skills'
    """Subdirectory of the resolved base directory where skills live.
    Ignored when ``skills_path_override`` is set."""

    skills_path_override: Optional[str] = None
    """Absolute path to the skills directory when it is not under the
    detection directory (e.g. Codex stores skills at ``~/.agents/skills/``).
    Tilde expansion is supported."""

    mcp_config_path: Optional[str] = None
    """Path to the MCP config file. Relative paths are resolved against
    the detection directory; tilde-prefixed paths are absolute. Leave
    ``None`` for agents that don't have JSON-based MCP config."""

    mcp_servers_key: str = 'mcpServers'
    """Top-level JSON key under which MCP servers are stored. Most
    agents use ``mcpServers``; OpenCode uses ``mcp``."""

    mcp_extra_config: Optional[dict] = None
    """Extra fields merged into the default MCP server entry (e.g.
    Kiro requires ``timeout`` and ``transport``). Used only when
    ``mcp_server_entry`` is not set."""

    mcp_server_entry: Optional[dict] = None
    """Complete MCP server entry, replacing the default schema. Use
    when an agent expects a different shape than ``{command, args}``
    (e.g. OpenCode expects ``{type, command: [...]}``)."""

    mcp_shell_command: Optional[list] = None
    """Argv for a CLI invocation that registers the AWS MCP server
    (e.g. ``codex mcp add ...``). Used for agents whose MCP config is
    not JSON. The wizard skips MCP setup if the executable is not on
    PATH."""

    def __post_init__(self):
        if self.mcp_extra_config is None:
            self.mcp_extra_config = {}

    @property
    def _display_skills_path(self):
        if self.skills_path_override:
            return self.skills_path_override.rstrip('/')
        base = self.detection_path.rstrip('/')
        if self.skills_dir:
            return f'{base}/{self.skills_dir}'
        return base

    @property
    def display_label(self):
        return f'{self.display_name} — {self._display_skills_path}'

    def resolved_override_dir(self):
        if not self.detection_path_env_override:
            return None
        env_value = os.environ.get(self.detection_path_env_override)
        if not env_value:
            return None
        expanded = os.path.expanduser(env_value).rstrip(os.sep)
        if not os.path.isdir(expanded):
            return None
        return expanded

    def resolve_base_dir(self):
        override = self.resolved_override_dir()
        if override is not None:
            return override
        expanded = os.path.expanduser(self.detection_path).rstrip(os.sep)
        if os.path.isdir(expanded):
            return expanded
        return None

    def detect(self):
        base_dir = self.resolve_base_dir()
        if base_dir is None:
            return None
        return DetectedAgent(self, base_dir)


class DetectedAgent:
    """An agent that was found installed on this machine."""

    def __init__(self, config, base_dir):
        self.config = config
        self.base_dir = base_dir

    @property
    def display_name(self):
        return self.config.display_name

    @property
    def display_label(self):
        return f'{self.display_name} — {collapse_home(self.skills_path)}'

    @property
    def skills_path(self):
        if self.config.skills_path_override:
            return os.path.expanduser(self.config.skills_path_override)
        return os.path.join(self.base_dir, self.config.skills_dir)

    @property
    def mcp_config_file(self):
        if self.config.mcp_config_path is None:
            return None
        if self.config.mcp_config_path.startswith('~'):
            override = self.config.resolved_override_dir()
            if override is not None:
                relative = self.config.mcp_config_path.removeprefix('~/')
                return os.path.join(override, relative)
            return os.path.expanduser(self.config.mcp_config_path)
        return os.path.join(self.base_dir, self.config.mcp_config_path)

    def get_installed_skills(self):
        pattern = os.path.join(
            self.skills_path,
            '*',
            SKILL_FILENAME,
        )
        results = []
        for skill_path in sorted(glob.glob(pattern)):
            skill_dir_path = os.path.dirname(skill_path)
            if not os.path.exists(
                os.path.join(skill_dir_path, SKILL_METADATA_FILENAME)
            ):
                continue
            skill_dir = os.path.basename(skill_dir_path)
            results.append(InstalledSkill(self, skill_dir, skill_path))
        return results

    def configure_mcp_server(self):
        """Configure the AWS MCP server entry for this agent.

        Returns a tuple of (action, detail) where action is a member of
        McpConfigureAction. detail is the config file path for JSON-based
        agents, the executable name for shell-based agents, or None when
        no MCP setup is wired up.
        """
        if self.config.mcp_shell_command is not None:
            return self._configure_via_shell()

        config_path = self.mcp_config_file
        if config_path is None:
            return McpConfigureAction.SKIPPED, None

        existing_config = self._read_mcp_config(config_path)
        servers = existing_config.setdefault(self.config.mcp_servers_key, {})

        if AWS_MCP_SERVER_KEY in servers:
            return McpConfigureAction.ALREADY_CONFIGURED, config_path

        if self.config.mcp_server_entry is not None:
            server_entry = self.config.mcp_server_entry
        else:
            server_entry = {
                **DEFAULT_MCP_SERVER_CONFIG,
                **self.config.mcp_extra_config,
            }
        servers[AWS_MCP_SERVER_KEY] = server_entry
        self._write_mcp_config(config_path, existing_config)
        return McpConfigureAction.CONFIGURED, config_path

    def _configure_via_shell(self):
        argv = list(self.config.mcp_shell_command)
        executable = argv[0]
        if shutil.which(executable) is None:
            return McpConfigureAction.SKIPPED, executable
        subprocess.run(argv, check=True)
        return McpConfigureAction.CONFIGURED, executable

    @staticmethod
    def _read_mcp_config(path):
        if not os.path.exists(path):
            return {}
        with open(path) as f:
            return json.load(f)

    @staticmethod
    def _write_mcp_config(path, config):
        os.makedirs(os.path.dirname(path), exist_ok=True)
        is_new = not os.path.exists(path)
        with open(path, 'w') as f:
            json.dump(config, f, indent=2)
            f.write('\n')
        # If we created the MCP config file, set permissions to 600, otherwise
        # the open call above preserves permissions for existing files
        if is_new:
            os.chmod(path, 0o600)


class InstalledSkill:
    """A skill found in an agent's skills directory."""

    def __init__(self, agent, name, path):
        self.agent = agent
        self.name = name
        self.path = path


AGENT_CONFIGS = [
    # https://docs.anthropic.com/en/docs/claude-code/mcp
    AgentConfig(
        id='claude-code',
        display_name='Claude Code',
        detection_path='~/.claude/',
        mcp_config_path='~/.claude.json',
        mcp_servers_key='mcpServers',
        detection_path_env_override='CLAUDE_CONFIG_DIR',
    ),
    # https://docs.cline.bot/mcp/mcp-overview
    # https://docs.cline.bot/customization/skills
    AgentConfig(
        id='cline',
        display_name='Cline',
        detection_path='~/.cline/',
        mcp_config_path='mcp.json',
        mcp_servers_key='mcpServers',
    ),
    # https://developers.openai.com/codex/skills
    # https://github.com/openai/codex/blob/main/codex-rs/cli/src/mcp_cmd.rs
    # Codex stores MCP config in TOML, not JSON. We shell out to its CLI
    # rather than add a TOML dependency.
    AgentConfig(
        id='codex',
        display_name='Codex',
        detection_path='~/.codex/',
        skills_path_override='~/.agents/skills/',
        detection_path_env_override='CODEX_HOME',
        mcp_shell_command=[
            'codex',
            'mcp',
            'add',
            AWS_MCP_SERVER_KEY,
            '--',
            'uvx',
            *_AWS_MCP_PROXY_ARGS,
        ],
    ),
    # https://docs.cursor.com/context/model-context-protocol
    # https://cursor.com/docs/skills
    AgentConfig(
        id='cursor',
        display_name='Cursor',
        detection_path='~/.cursor/',
        mcp_config_path='mcp.json',
        mcp_servers_key='mcpServers',
    ),
    # https://geminicli.com/docs/cli/tutorials/mcp-setup/
    # https://geminicli.com/docs/cli/skills/
    AgentConfig(
        id='gemini-cli',
        display_name='Gemini CLI',
        detection_path='~/.gemini/',
        skills_path_override='~/.agents/skills/',
        mcp_config_path='settings.json',
        mcp_servers_key='mcpServers',
    ),
    # https://kiro.dev/docs/mcp/configuration/
    AgentConfig(
        id='kiro',
        display_name='Kiro',
        detection_path='~/.kiro/',
        mcp_config_path='settings/mcp.json',
        mcp_servers_key='mcpServers',
        mcp_extra_config={'timeout': 100000, 'transport': 'stdio'},
    ),
    # https://openclaw-openclaw.mintlify.app/configuration
    # OpenClaw does not document MCP support — install skills only.
    AgentConfig(
        id='openclaw',
        display_name='OpenClaw',
        detection_path='~/.openclaw/',
    ),
    # https://docs.opencode.ai/docs/mcp-servers
    # https://docs.opencode.ai/docs/skills
    AgentConfig(
        id='opencode',
        display_name='OpenCode',
        detection_path='~/.config/opencode/',
        skills_path_override='~/.agents/skills/',
        mcp_config_path='opencode.json',
        mcp_servers_key='mcp',
        mcp_server_entry={
            'type': 'local',
            'command': ['uvx', *_AWS_MCP_PROXY_ARGS],
        },
    ),
    # https://github.com/badlogic/pi-mono/blob/main/packages/coding-agent/docs/settings.md
    # Pi does not document MCP support — install skills only.
    AgentConfig(
        id='pi',
        display_name='Pi',
        detection_path='~/.pi/agent/',
    ),
    # https://docs.windsurf.com/plugins/cascade/mcp
    # https://docs.windsurf.com/windsurf/cascade/skills
    AgentConfig(
        id='windsurf',
        display_name='Windsurf',
        detection_path='~/.codeium/windsurf/',
        skills_path_override='~/.agents/skills/',
        mcp_config_path='~/.codeium/mcp_config.json',
        mcp_servers_key='mcpServers',
    ),
]


def _build_universal_row(configs):
    universal_path = os.path.expanduser('~/.agents/skills/').rstrip(os.sep)
    consumers = [
        c.display_name
        for c in configs
        if c.skills_path_override
        and os.path.expanduser(c.skills_path_override).rstrip(os.sep)
        == universal_path
    ]
    return AgentConfig(
        id=UNIVERSAL_ROW_ID,
        display_name=f'Universal ({", ".join(consumers)})',
        detection_path='~/.agents',
    )


AGENT_CONFIGS.append(_build_universal_row(AGENT_CONFIGS))


def universal_first(items, get_id=lambda a: a.config.id):
    return sorted(
        items,
        key=lambda i: 0 if get_id(i) == UNIVERSAL_ROW_ID else 1,
    )


def get_detected_agents(agent_configs=None):
    if agent_configs is None:
        agent_configs = AGENT_CONFIGS
    detected = []
    for config in agent_configs:
        agent = config.detect()
        if agent is not None:
            detected.append(agent)
    return detected



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
from unittest.mock import patch

from awscli.customizations.agenttoolkit.agents import (
    AGENT_CONFIGS,
    AgentConfig,
    DetectedAgent,
    McpConfigureAction,
    get_detected_agents,
)
from awscli.testutils import skip_if_windows
from tests.unit.customizations.agenttoolkit.utils import (
    make_config,
    make_skill,
)


def test_detect_when_installed(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    config = make_config(tmp_path)
    agent = config.detect()
    assert isinstance(agent, DetectedAgent)
    assert agent.display_name == 'Test Agent'
    assert agent.base_dir == str(tmp_path / '.test-agent')


def test_detect_when_not_installed(tmp_path):
    config = make_config(tmp_path)
    assert config.detect() is None


def test_detect_env_override(tmp_path, monkeypatch):
    (tmp_path / '.test-agent').mkdir()
    override_dir = tmp_path / '.custom-location'
    override_dir.mkdir()
    config = make_config(
        tmp_path,
        detection_path_env_override='TEST_CONFIG_DIR',
    )
    monkeypatch.setenv('TEST_CONFIG_DIR', str(override_dir))
    agent = config.detect()
    assert agent.base_dir == str(override_dir)


def test_detect_env_override_falls_back(tmp_path, monkeypatch):
    (tmp_path / '.test-agent').mkdir()
    config = make_config(
        tmp_path,
        detection_path_env_override='TEST_CONFIG_DIR',
    )
    monkeypatch.setenv('TEST_CONFIG_DIR', str(tmp_path / 'nonexistent'))
    agent = config.detect()
    assert agent.base_dir == str(tmp_path / '.test-agent')


def test_get_installed_skills_finds_aws_skills(tmp_path):
    make_skill(tmp_path, '.test-agent', 'aws-serverless')
    make_skill(tmp_path, '.test-agent', 'aws-databases')
    config = make_config(tmp_path)
    agent = config.detect()
    skills = agent.get_installed_skills()
    assert len(skills) == 2
    assert skills[0].name == 'aws-databases'
    assert skills[0].agent.display_name == 'Test Agent'
    assert skills[1].name == 'aws-serverless'
    assert skills[1].path.endswith(os.path.join('aws-serverless', 'SKILL.md'))


def test_skills_path_override(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    override_dir = tmp_path / 'shared' / 'skills'
    config = make_config(
        tmp_path,
        skills_path_override=str(override_dir),
    )
    agent = config.detect()
    assert agent.skills_path == str(override_dir)


def test_universal_row_display_name_lists_consumers():
    from awscli.customizations.agenttoolkit.agents import AGENT_CONFIGS

    universal = next(c for c in AGENT_CONFIGS if c.id == 'universal')
    assert universal.detection_path == '~/.agents'
    for config in AGENT_CONFIGS:
        if config.skills_path_override and config.skills_path_override.rstrip(
            '/'
        ).endswith('.agents/skills'):
            assert config.display_name in universal.display_name


def test_get_installed_skills_excludes_unmarked(tmp_path):
    make_skill(tmp_path, '.test-agent', 'aws-installed', with_marker=True)
    make_skill(tmp_path, '.test-agent', 'hand-rolled', with_marker=False)
    config = make_config(tmp_path)
    skills = config.detect().get_installed_skills()
    names = [s.name for s in skills]
    assert names == ['aws-installed']


def test_get_installed_skills_empty(tmp_path):
    (tmp_path / '.test-agent' / 'skills').mkdir(parents=True)
    config = make_config(tmp_path)
    assert config.detect().get_installed_skills() == []


def test_get_detected_agents(tmp_path):
    (tmp_path / '.kiro').mkdir()
    test_configs = [
        AgentConfig(
            id='kiro',
            display_name='Kiro',
            detection_path=str(tmp_path / '.kiro'),
        ),
        AgentConfig(
            id='cursor',
            display_name='Cursor',
            detection_path=str(tmp_path / '.cursor'),
        ),
    ]
    detected = get_detected_agents(agent_configs=test_configs)
    assert len(detected) == 1
    assert detected[0].display_name == 'Kiro'


def test_mcp_config_path_honors_detection_env_override(tmp_path, monkeypatch):
    (tmp_path / '.test-agent').mkdir()
    override_dir = tmp_path / '.custom-location'
    override_dir.mkdir()
    config = make_config(
        tmp_path,
        detection_path_env_override='TEST_CONFIG_DIR',
        mcp_config_path='~/.test-agent.json',
    )
    monkeypatch.setenv('TEST_CONFIG_DIR', str(override_dir))
    agent = config.detect()
    assert agent.mcp_config_file == str(override_dir / '.test-agent.json')


def test_mcp_config_path_falls_back_when_env_override_dir_missing(
    tmp_path, monkeypatch
):
    (tmp_path / '.test-agent').mkdir()
    config = make_config(
        tmp_path,
        detection_path_env_override='TEST_CONFIG_DIR',
        mcp_config_path='~/.test-agent.json',
    )
    monkeypatch.setenv('TEST_CONFIG_DIR', str(tmp_path / 'nonexistent'))
    agent = config.detect()
    assert agent.mcp_config_file == os.path.expanduser('~/.test-agent.json')


def test_mcp_config_path_falls_back_to_home_without_env_override(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    config = make_config(tmp_path, mcp_config_path='~/.test-agent.json')
    agent = config.detect()
    assert agent.mcp_config_file == os.path.expanduser('~/.test-agent.json')


def test_configure_mcp_new_file(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    agent = make_config(tmp_path, mcp_config_path='mcp.json').detect()
    action, path = agent.configure_mcp_server()
    assert action is McpConfigureAction.CONFIGURED
    data = json.loads(open(path).read())
    assert data['mcpServers']['aws-mcp']['command'] == 'uvx'
    assert 'mcp-proxy-for-aws@latest' in data['mcpServers']['aws-mcp']['args']


def test_configure_mcp_already_configured(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    agent = make_config(tmp_path, mcp_config_path='mcp.json').detect()
    mcp_path = tmp_path / '.test-agent' / 'mcp.json'
    mcp_path.write_text(
        json.dumps({'mcpServers': {'aws-mcp': {'command': 'custom'}}})
    )
    action, _ = agent.configure_mcp_server()
    assert action is McpConfigureAction.ALREADY_CONFIGURED
    data = json.loads(mcp_path.read_text())
    assert data['mcpServers']['aws-mcp']['command'] == 'custom'


def test_configure_mcp_preserves_other_servers(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    agent = make_config(tmp_path, mcp_config_path='mcp.json').detect()
    mcp_path = tmp_path / '.test-agent' / 'mcp.json'
    mcp_path.write_text(
        json.dumps({'mcpServers': {'other': {'command': 'foo'}}})
    )
    agent.configure_mcp_server()
    data = json.loads(mcp_path.read_text())
    assert data['mcpServers']['other']['command'] == 'foo'
    assert 'aws-mcp' in data['mcpServers']


def test_configure_mcp_extra_config_merged(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    agent = make_config(
        tmp_path,
        mcp_config_path='mcp.json',
        mcp_extra_config={'timeout': 100000},
    ).detect()
    agent.configure_mcp_server()
    mcp_path = tmp_path / '.test-agent' / 'mcp.json'
    data = json.loads(mcp_path.read_text())
    entry = data['mcpServers']['aws-mcp']
    assert entry['timeout'] == 100000
    assert entry['command'] == 'uvx'


def test_configure_mcp_server_entry_overrides_default(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    custom_entry = {
        'type': 'local',
        'command': ['uvx', 'something@latest'],
    }
    agent = make_config(
        tmp_path,
        mcp_config_path='mcp.json',
        mcp_servers_key='mcp',
        mcp_server_entry=custom_entry,
    ).detect()
    agent.configure_mcp_server()
    mcp_path = tmp_path / '.test-agent' / 'mcp.json'
    data = json.loads(mcp_path.read_text())
    entry = data['mcp']['aws-mcp']
    assert entry == custom_entry
    assert 'args' not in entry


def test_configure_mcp_shell_command_runs_when_executable_present(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    config = make_config(
        tmp_path,
        mcp_shell_command=['some-cli', 'mcp', 'add', 'aws-mcp'],
    )
    agent = config.detect()
    with (
        patch(
            'awscli.customizations.agenttoolkit.agents.shutil.which',
            return_value='/usr/local/bin/some-cli',
        ),
        patch(
            'awscli.customizations.agenttoolkit.agents.subprocess.run'
        ) as run_mock,
    ):
        action, detail = agent.configure_mcp_server()
    assert action is McpConfigureAction.CONFIGURED
    assert detail == 'some-cli'
    run_mock.assert_called_once_with(
        ['some-cli', 'mcp', 'add', 'aws-mcp'], check=True
    )


def test_configure_mcp_shell_command_skipped_when_executable_missing(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    config = make_config(
        tmp_path,
        mcp_shell_command=['missing-cli', 'mcp', 'add', 'aws-mcp'],
    )
    agent = config.detect()
    with (
        patch(
            'awscli.customizations.agenttoolkit.agents.shutil.which',
            return_value=None,
        ),
        patch(
            'awscli.customizations.agenttoolkit.agents.subprocess.run'
        ) as run_mock,
    ):
        action, detail = agent.configure_mcp_server()
    assert action is McpConfigureAction.SKIPPED
    assert detail == 'missing-cli'
    run_mock.assert_not_called()


def test_configure_mcp_creates_parent_directories(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    agent = make_config(
        tmp_path, mcp_config_path='nested/dir/mcp.json'
    ).detect()
    agent.configure_mcp_server()
    assert (tmp_path / '.test-agent' / 'nested' / 'dir' / 'mcp.json').exists()


@skip_if_windows("POSIX file permissions are not enforced on Windows.")
def test_configure_mcp_new_file_gets_0600_permissions(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    agent = make_config(tmp_path, mcp_config_path='mcp.json').detect()
    agent.configure_mcp_server()
    mcp_path = tmp_path / '.test-agent' / 'mcp.json'
    assert oct(mcp_path.stat().st_mode & 0o777) == oct(0o600)


@skip_if_windows("POSIX file permissions are not enforced on Windows.")
def test_configure_mcp_preserves_existing_permissions(tmp_path):
    (tmp_path / '.test-agent').mkdir()
    mcp_path = tmp_path / '.test-agent' / 'mcp.json'
    mcp_path.write_text(json.dumps({'mcpServers': {}}))
    os.chmod(mcp_path, 0o644)
    agent = make_config(tmp_path, mcp_config_path='mcp.json').detect()
    agent.configure_mcp_server()
    assert oct(mcp_path.stat().st_mode & 0o777) == oct(0o644)


def test_agent_configs_sorted_alphabetically():
    ids = [c.id for c in AGENT_CONFIGS if c.id != 'universal']
    assert ids == sorted(ids)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import sys

from awscli.customizations.agenttoolkit.agents import AGENT_CONFIGS
from awscli.customizations.agenttoolkit.utils import (
    AGENT_ARG,
    SKILL_NAME_ARG,
    create_client,
    get_skill_download,
    install_skill,
    read_installed_version,
    resolve_agents,
    resolve_latest_version,
)
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError


class UpdateSkillCommand(BasicCommand):
    NAME = 'update-skill'
    DESCRIPTION = (
        'Update an installed AWS skill to the latest version. '
        'Compares the locally installed version against the available skills '
        'and downloads the newer version if available. By default the skill is '
        'updated for all detected agents, use ``--agent`` to update the skill '
        'for only a specific tool.'
    )
    ARG_TABLE = [
        SKILL_NAME_ARG,
        AGENT_ARG,
    ]

    def __init__(self, session, stream=None, client=None, agent_configs=None):
        super().__init__(session)
        if stream is None:
            stream = sys.stdout
        self._stream = stream
        self._client = client
        if agent_configs is None:
            agent_configs = AGENT_CONFIGS
        self._agent_configs = agent_configs

    def _run_main(self, parsed_args, parsed_globals):
        skill_name = parsed_args.skill_name
        agent_filter = getattr(parsed_args, 'agent', None)

        agents = resolve_agents(agent_filter, self._agent_configs)
        if not agents:
            raise ParamValidationError('No supported AI coding agents found.')

        installed_agents = [
            agent
            for agent in agents
            if any(
                skill.name == skill_name
                for skill in agent.get_installed_skills()
            )
        ]
        if not installed_agents:
            raise ParamValidationError(
                f'Skill "{skill_name}" is not installed.'
            )

        client = self._client or create_client(self._session, parsed_globals)
        remote_version = resolve_latest_version(client, skill_name)

        outdated = []
        for agent in installed_agents:
            skill_dir = os.path.join(agent.skills_path, skill_name)
            local_version = read_installed_version(skill_dir)
            if local_version != remote_version:
                outdated.append(agent)

        if not outdated:
            self._stream.write(
                f'{skill_name} is already up to date ({remote_version}).\n'
            )
            return 0

        zip_bytes, checksum, version = get_skill_download(
            client, skill_name, version=remote_version
        )
        install_skill(
            skill_name,
            version,
            zip_bytes,
            checksum,
            outdated,
            self._stream,
            action='Updated',
            overwrite_existing=True,
        )
        return 0



# Copyright 2026 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.emr.ssh import SSH, Get, Put, Socks
from awscli.testutils import mock
from tests.unit.customizations.emr import (
    EMRBaseAWSCommandParamsTest as BaseAWSCommandParamsTest,
)

CLUSTER_ID = "j-TESTCLUSTER123"
KEY_PAIR_FILE = "/tmp/test_key.pem"
MASTER_DNS = "ec2-1-2-3-4.us-west-2.compute.amazonaws.com"


class TestSSHHostKeyVerification(BaseAWSCommandParamsTest):
    """Verify all EMR SSH commands use StrictHostKeyChecking=accept-new
    when OpenSSH supports it."""

    def _get_ssh_command(self, mock_call):
        self.assertTrue(mock_call.called)
        return mock_call.call_args[0][0]

    def _assert_accept_new(self, command):
        joined = ' '.join(command)
        self.assertIn('StrictHostKeyChecking=accept-new', joined)
        self.assertNotIn('StrictHostKeyChecking=no', joined)

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.ssh._supports_accept_new',
                return_value=True)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_ssh_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/ssh')
    def test_socks_uses_accept_new(self, mock_which, mock_validate,
                                   mock_dns, mock_probe, mock_call):
        self.run_cmd(
            f'emr socks --cluster-id {CLUSTER_ID} '
            f'--key-pair-file {KEY_PAIR_FILE}', 0)
        command = self._get_ssh_command(mock_call)
        self._assert_accept_new(command)

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.ssh._supports_accept_new',
                return_value=True)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_ssh_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/ssh')
    def test_ssh_uses_accept_new(self, mock_which, mock_validate,
                                 mock_dns, mock_probe, mock_call):
        self.run_cmd(
            f'emr ssh --cluster-id {CLUSTER_ID} '
            f'--key-pair-file {KEY_PAIR_FILE}', 0)
        command = self._get_ssh_command(mock_call)
        self._assert_accept_new(command)

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.ssh._supports_accept_new',
                return_value=True)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_scp_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/scp')
    def test_put_uses_accept_new(self, mock_which, mock_validate,
                                 mock_dns, mock_probe, mock_call):
        self.run_cmd(
            f'emr put --cluster-id {CLUSTER_ID} '
            f'--key-pair-file {KEY_PAIR_FILE} --src /tmp/local.txt', 0)
        command = self._get_ssh_command(mock_call)
        self._assert_accept_new(command)

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.ssh._supports_accept_new',
                return_value=True)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_scp_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/scp')
    def test_get_uses_accept_new(self, mock_which, mock_validate,
                                 mock_dns, mock_probe, mock_call):
        self.run_cmd(
            f'emr get --cluster-id {CLUSTER_ID} '
            f'--key-pair-file {KEY_PAIR_FILE} --src /remote/file.txt', 0)
        command = self._get_ssh_command(mock_call)
        self._assert_accept_new(command)


class TestSSHFallbackOnOldOpenSSH(BaseAWSCommandParamsTest):
    """Verify fallback to StrictHostKeyChecking=no when accept-new
    is not supported, with a warning."""

    def _get_ssh_command(self, mock_call):
        self.assertTrue(mock_call.called)
        return mock_call.call_args[0][0]

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.ssh._supports_accept_new',
                return_value=False)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_ssh_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/ssh')
    def test_falls_back_to_no_with_warning(self, mock_which, mock_validate,
                                           mock_dns, mock_probe, mock_call):
        _, stderr, _ = self.run_cmd(
            f'emr ssh --cluster-id {CLUSTER_ID} '
            f'--key-pair-file {KEY_PAIR_FILE}', 0)
        command = self._get_ssh_command(mock_call)
        joined = ' '.join(command)
        self.assertIn('StrictHostKeyChecking=no', joined)
        self.assertNotIn('StrictHostKeyChecking=accept-new', joined)
        self.assertIn('Upgrade to OpenSSH 7.6+', stderr)


class TestSSHOptionsPassthrough(BaseAWSCommandParamsTest):
    """Verify --ssh-options are passed through and can override defaults."""

    def _get_ssh_command(self, mock_call):
        self.assertTrue(mock_call.called)
        return mock_call.call_args[0][0]

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_ssh_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/ssh')
    def test_ssh_options_override_strict_host_key(self, mock_which,
                                                  mock_validate,
                                                  mock_dns, mock_call):
        self.run_cmd(
            f'emr ssh --cluster-id {CLUSTER_ID} --key-pair-file {KEY_PAIR_FILE} '
            '--ssh-options StrictHostKeyChecking=no', 0)
        command = self._get_ssh_command(mock_call)
        joined = ' '.join(command)
        self.assertIn('StrictHostKeyChecking=no', joined)
        self.assertNotIn('StrictHostKeyChecking=accept-new', joined)

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.ssh._supports_accept_new',
                return_value=True)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_ssh_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/ssh')
    def test_ssh_options_additional_options(self, mock_which, mock_validate,
                                           mock_dns, mock_probe, mock_call):
        self.run_cmd(
            f'emr ssh --cluster-id {CLUSTER_ID} --key-pair-file {KEY_PAIR_FILE} '
            '--ssh-options ConnectTimeout=30', 0)
        command = self._get_ssh_command(mock_call)
        joined = ' '.join(command)
        self.assertIn('StrictHostKeyChecking=accept-new', joined)
        self.assertIn('ConnectTimeout=30', joined)

    @mock.patch('subprocess.call', return_value=0)
    @mock.patch('awscli.customizations.emr.sshutils.validate_and_find_master_dns',
                return_value=MASTER_DNS)
    @mock.patch('awscli.customizations.emr.sshutils.validate_scp_with_key_file')
    @mock.patch('awscli.customizations.emr.emrutils.which', return_value='/usr/bin/scp')
    def test_put_options_override(self, mock_which, mock_validate,
                                  mock_dns, mock_call):
        self.run_cmd(
            f'emr put --cluster-id {CLUSTER_ID} --key-pair-file {KEY_PAIR_FILE} '
            '--src /tmp/f.txt '
            '--ssh-options StrictHostKeyChecking=no', 0)
        command = self._get_ssh_command(mock_call)
        joined = ' '.join(command)
        self.assertIn('StrictHostKeyChecking=no', joined)
        self.assertNotIn('StrictHostKeyChecking=accept-new', joined)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import shutil
import sys

from awscli.customizations.agenttoolkit.agents import (
    AGENT_CONFIGS,
    universal_first,
)
from awscli.customizations.agenttoolkit.utils import (
    AGENT_ARG,
    SKILL_NAME_ARG,
    resolve_agents,
)
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError


class RemoveCommand(BasicCommand):
    NAME = 'remove-skill'
    DESCRIPTION = (
        'Remove a previously installed AWS skill from detected agents. '
        'By default the skill is removed from all detected agents, use '
        '``--agent`` to remove from a specific tool only.'
    )
    ARG_TABLE = [
        SKILL_NAME_ARG,
        AGENT_ARG,
    ]

    def __init__(self, session, stream=None, agent_configs=None):
        super().__init__(session)
        if stream is None:
            stream = sys.stdout
        self._stream = stream
        if agent_configs is None:
            agent_configs = AGENT_CONFIGS
        self._agent_configs = agent_configs

    def _run_main(self, parsed_args, parsed_globals):
        skill_name = parsed_args.skill_name
        agent_filter = getattr(parsed_args, 'agent', None)

        agents = resolve_agents(agent_filter, self._agent_configs)

        matches = []
        for agent in agents:
            for skill in agent.get_installed_skills():
                if skill.name == skill_name:
                    matches.append(skill)

        if not matches:
            raise ParamValidationError(
                f'Skill "{skill_name}" is not installed.'
            )

        removed_paths = set()
        ordered = universal_first(matches, get_id=lambda s: s.agent.config.id)
        for skill in ordered:
            skill_dir = os.path.dirname(skill.path)
            if skill_dir in removed_paths:
                continue
            shutil.rmtree(skill_dir)
            removed_paths.add(skill_dir)
            self._stream.write(
                f'Removed {skill.name} from {skill.agent.display_label}.\n'
            )
        return 0



# Copyright 2026 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestSigninOAuth2Alias(BaseAWSCommandParamsTest):
    def test_create_oauth2_token_alias(self):
        command_template = (
            'signin %s --token-input '
            'clientId=c,grantType=authorization_code'
        )
        self.run_cmd(command_template % 'create-o-auth2-token', expected_rc=0)
        self.run_cmd(command_template % 'create-oauth2-token', expected_rc=0)

    def test_create_oauth2_token_with_iam_rename(self):
        args = '--grant-type client_credentials --resource r'
        self.run_cmd(
            f'signin create-oauth2-token-with-iam {args}', expected_rc=0
        )
        self.run_cmd(
            f'signin create-o-auth2-token-with-iam {args}', expected_rc=252
        )

    def test_introspect_oauth2_token_with_iam_rename(self):
        args = '--token t'
        self.run_cmd(
            f'signin introspect-oauth2-token-with-iam {args}', expected_rc=0
        )
        self.run_cmd(
            f'signin introspect-o-auth2-token-with-iam {args}', expected_rc=252
        )

    def test_revoke_oauth2_token_with_iam_rename(self):
        args = '--token t'
        self.run_cmd(
            f'signin revoke-oauth2-token-with-iam {args}', expected_rc=0
        )
        self.run_cmd(
            f'signin revoke-o-auth2-token-with-iam {args}', expected_rc=252
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys

from awscli.customizations.agenttoolkit.agents import AGENT_CONFIGS
from awscli.customizations.agenttoolkit.utils import (
    AGENT_ARG,
    SKILL_NAME_ARG,
    SKILL_VERSION_ARG,
    create_client,
    get_skill_download,
    install_skill,
    resolve_agents,
)
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError


class AddSkillCommand(BasicCommand):
    NAME = 'add-skill'
    DESCRIPTION = (
        'Download and install an AWS skill to detected AI coding agents. '
        'By default the latest version is installed globally to all detected '
        'agents. Use ``--agent`` to target a specific tool, or '
        '``--skill-version`` to pin a specific version.'
    )
    ARG_TABLE = [
        SKILL_NAME_ARG,
        SKILL_VERSION_ARG,
        AGENT_ARG,
    ]

    def __init__(self, session, stream=None, client=None, agent_configs=None):
        super().__init__(session)
        if stream is None:
            stream = sys.stdout
        self._stream = stream
        self._client = client
        if agent_configs is None:
            agent_configs = AGENT_CONFIGS
        self._agent_configs = agent_configs

    def _run_main(self, parsed_args, parsed_globals):
        skill_name = parsed_args.skill_name
        version = getattr(parsed_args, 'skill_version', None)
        agent_filter = getattr(parsed_args, 'agent', None)

        agents = resolve_agents(agent_filter, self._agent_configs)
        if not agents:
            raise ParamValidationError('No supported AI coding agents found.')

        client = self._client or create_client(self._session, parsed_globals)
        zip_bytes, checksum, version = get_skill_download(
            client, skill_name, version=version
        )

        install_skill(
            skill_name, version, zip_bytes, checksum, agents, self._stream
        )
        return 0



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.agenttoolkit.agents import universal_first
from awscli.customizations.agenttoolkit.utils import (
    AGENT_ARG,
    resolve_agents,
)
from awscli.customizations.commands import BasicCommand
from awscli.formatter import get_formatter
from awscli.utils import OutputStreamFactory


class ListInstalledSkillsCommand(BasicCommand):
    NAME = 'list-installed-skills'
    DESCRIPTION = (
        'List AWS skills that were previously installed by the '
        '``aws agent-toolkit`` commands. Shows the skill name, agent, and file '
        'path for each installation. By default it lists skills from all '
        'detected agents, use ``--agent`` to filter results to a specific tool.'
    )
    ARG_TABLE = [AGENT_ARG]

    def __init__(
        self, session, agent_configs=None, output_stream_factory=None
    ):
        super().__init__(session)
        self._agent_configs = agent_configs
        if output_stream_factory is None:
            output_stream_factory = OutputStreamFactory(session)
        self._output_stream_factory = output_stream_factory

    def _run_main(self, parsed_args, parsed_globals):
        agent_filter = getattr(parsed_args, 'agent', None)
        agents = resolve_agents(agent_filter, self._agent_configs)

        seen_paths = set()
        all_skills = []
        for agent in universal_first(agents):
            for skill in agent.get_installed_skills():
                if skill.path in seen_paths:
                    continue
                seen_paths.add(skill.path)
                all_skills.append(skill)

        result = {
            'skills': [
                {
                    'agent': skill.agent.display_name,
                    'name': skill.name,
                    'path': skill.path,
                }
                for skill in all_skills
            ]
        }

        output = parsed_globals.output
        if output is None:
            output = self._session.get_config_variable('output')
        formatter = get_formatter(output, parsed_globals)
        with self._output_stream_factory.get_output_stream() as stream:
            formatter(self.NAME, result, stream=stream)
            stream.write('\n')
        return 0



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import sqlite3
from unittest.mock import MagicMock, patch

import pytest
from botocore.exceptions import MD5UnavailableError
from botocore.session import Session

from awscli.clidriver import create_clidriver
from awscli.telemetry import (
    CLISessionData,
    CLISessionDatabaseConnection,
    CLISessionDatabaseReader,
    CLISessionDatabaseSweeper,
    CLISessionDatabaseWriter,
    CLISessionGenerator,
    CLISessionOrchestrator,
    is_telemetry_disabled,
    register_session_id_event,
)
from tests.markers import skip_if_windows


@pytest.fixture
def session_conn():
    conn = CLISessionDatabaseConnection(
        connection=sqlite3.connect(
            # Use an in-memory db for testing.
            ':memory:',
            check_same_thread=False,
            isolation_level=None,
        ),
    )
    # Write an initial record.
    conn.execute(
        """
            INSERT OR REPLACE INTO session (
                key, session_id, timestamp
            ) VALUES ('first_key', 'first_id', 5555555555)
        """
    )
    # Overwrite host id with deterministic value for testing.
    conn.execute(
        """
            INSERT OR REPLACE INTO host_id (
                key, id
            ) VALUES (0, 'my-hostname')
        """
    )
    return conn


@pytest.fixture
def session_writer(session_conn):
    return CLISessionDatabaseWriter(session_conn)


@pytest.fixture
def session_reader(session_conn):
    return CLISessionDatabaseReader(session_conn)


@pytest.fixture
def session_sweeper(session_conn):
    return CLISessionDatabaseSweeper(session_conn)


@pytest.fixture
def session_generator():
    return CLISessionGenerator()


@pytest.fixture
def expired_data(session_writer, session_reader, session_sweeper):
    # Write an expired record.
    session_writer.write(
        CLISessionData(
            key='expired_key',
            session_id='expired_id',
            timestamp=1000000000,
        )
    )
    # Ensure expired record exists.
    assert session_reader.read('expired_key') is not None
    yield
    # Ensure cleanup after test is run.
    session_sweeper.sweep(1000000001)


class TestCLISessionDatabaseConnection:
    def test_ensure_database_setup(self, session_conn):
        cursor = session_conn.execute(
            """
                SELECT name
                FROM sqlite_master
                WHERE type='table';
            """
        )
        assert cursor.fetchall() == [('session',), ('host_id',)]

    def test_creates_database_when_cache_dir_does_not_exist(self, tmp_path):
        # When the cache directory doesn't exist, the connection should still
        # be established successfully.
        nonexistent_dir = tmp_path / 'nonexistent' / 'nested' / 'cache'
        assert not nonexistent_dir.exists()
        conn = CLISessionDatabaseConnection(cache_dir=nonexistent_dir)
        assert nonexistent_dir.exists()
        assert (nonexistent_dir / 'session.db').exists()
        # Verify the database is functional.
        writer = CLISessionDatabaseWriter(conn)
        reader = CLISessionDatabaseReader(conn)
        writer.write(CLISessionData('key', 'sid', 1000000000))
        assert reader.read('key').session_id == 'sid'

    def test_timeout_does_not_raise_exception(self, session_conn):
        test_query = """
            SELECT name
            FROM sqlite_master
            WHERE type='table'
            AND name='session';
        """

        class FakeConnection(sqlite3.Connection):
            def execute(self, query, *parameters):
                # Simulate timeout by always raising.
                if query == test_query:
                    raise sqlite3.OperationalError()
                # Mock host id count query.
                cur = MagicMock()
                cur.fetchone.return_value = (1,)
                return cur

        fake_conn = CLISessionDatabaseConnection(FakeConnection(":memory:"))
        cursor = fake_conn.execute(test_query)
        assert cursor.fetchall() == []


class TestCLISessionDatabaseWriter:
    def test_write(self, session_writer, session_reader, session_sweeper):
        session_writer.write(
            CLISessionData(
                key='new-key',
                session_id='new-id',
                timestamp=1000000000,
            )
        )
        session_data = session_reader.read('new-key')
        assert session_data.key == 'new-key'
        assert session_data.session_id == 'new-id'
        assert session_data.timestamp == 1000000000
        session_sweeper.sweep(1000000001)


class TestCLISessionDatabaseReader:
    def test_read(self, session_reader):
        session_data = session_reader.read('first_key')
        assert session_data.key == 'first_key'
        assert session_data.session_id == 'first_id'
        assert session_data.timestamp == 5555555555

    def test_read_nonexistent_record(self, session_reader):
        session_data = session_reader.read('bad_key')
        assert session_data is None

    def test_read_host_id(self, session_reader):
        host_id = session_reader.read_host_id()
        assert host_id == 'my-hostname'


class TestCLISessionDatabaseSweeper:
    def test_sweep(self, expired_data, session_reader, session_sweeper):
        session_sweeper.sweep(1000000001)
        swept_data = session_reader.read('expired_key')
        assert swept_data is None

    def test_sweep_not_expired(
        self, expired_data, session_reader, session_sweeper
    ):
        session_sweeper.sweep(1000000000)
        swept_data = session_reader.read('expired_key')
        assert swept_data is not None

    def test_sweep_never_raises(self, session_sweeper):
        # Normally this would raise `sqlite3.ProgrammingError`,
        # but the `sweep` method catches bare exceptions.
        session_sweeper.sweep({'bad': 'input'})


class TestCLISessionGenerator:
    def test_generate_session_id(self, session_generator):
        session_id = session_generator.generate_session_id(
            'my-hostname',
            'my-tty',
            1000000000,
        )
        assert session_id == 'd949713b13ee'

    def test_generate_cache_key(self, session_generator):
        cache_key = session_generator.generate_cache_key(
            'my-hostname',
            'my-tty',
        )
        assert cache_key == 'b1ca2be0ffac'

    @patch('awscli.telemetry.get_md5')
    def test_checksum_fips_fallback(self, patched_get_md5, session_generator):
        patched_get_md5.side_effect = MD5UnavailableError()
        session_id = session_generator.generate_session_id(
            'my-hostname',
            'my-tty',
            1000000000,
        )
        assert session_id == '183b154db015'


@skip_if_windows
@patch('sys.stdin')
@patch('time.time', return_value=5555555555)
@patch('os.ttyname', return_value='my-tty')
class TestCLISessionOrchestrator:
    def test_session_id_gets_cached(
        self,
        patched_tty_name,
        patched_time,
        patched_stdin,
        session_sweeper,
        session_generator,
        session_reader,
        session_writer,
    ):
        patched_stdin.fileno.return_value = None
        orchestrator = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        assert orchestrator.session_id == '881cea8546fa'

        session_data = session_reader.read(orchestrator.cache_key)
        assert session_data.key == orchestrator.cache_key
        assert session_data.session_id == orchestrator.session_id
        assert session_data.timestamp == 5555555555

    def test_cached_session_id_updated_if_expired(
        self,
        patched_tty_name,
        patched_time,
        patched_stdin,
        session_sweeper,
        session_generator,
        session_reader,
        session_writer,
    ):
        patched_stdin.fileno.return_value = None

        # First, generate and cache a session id.
        orchestrator_1 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_1 = orchestrator_1.session_id
        session_data_1 = session_reader.read(orchestrator_1.cache_key)
        assert session_data_1.session_id == session_id_1

        # Update the timestamp and get the new session id.
        patched_time.return_value = 7777777777
        orchestrator_2 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_2 = orchestrator_2.session_id
        session_data_2 = session_reader.read(orchestrator_2.cache_key)

        # Cache key should be the same.
        assert session_data_2.key == session_data_1.key
        # Session id and timestamp should be updated.
        assert session_data_2.session_id == session_id_2
        assert session_data_2.session_id != session_data_1.session_id
        assert session_data_2.timestamp == 7777777777
        assert session_data_2.timestamp != session_data_1.timestamp

    def test_cached_session_id_not_updated_if_valid(
        self,
        patched_tty_name,
        patched_time,
        patched_stdin,
        session_sweeper,
        session_generator,
        session_reader,
        session_writer,
    ):
        patched_stdin.fileno.return_value = None

        # First, generate and cache a session id.
        orchestrator_1 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_1 = orchestrator_1.session_id
        session_data_1 = session_reader.read(orchestrator_1.cache_key)
        assert session_data_1.session_id == session_id_1

        # Update the timestamp.
        patched_time.return_value = 5555555556
        orchestrator_2 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_2 = orchestrator_2.session_id
        session_data_2 = session_reader.read(orchestrator_2.cache_key)

        # Cache key should be the same.
        assert session_data_2.key == session_data_1.key
        # Session id should not be updated.
        assert session_data_2.session_id == session_id_2
        assert session_data_2.session_id == session_data_1.session_id
        # Only timestamp should be updated.
        assert session_data_2.timestamp == 5555555556
        assert session_data_2.timestamp != session_data_1.timestamp


def test_register_session_id_event_injects_sid_on_before_create_client():
    session = MagicMock(Session)
    session.user_agent_extra = 'md/installer#source'
    event_emitter = MagicMock()
    session.get_component.return_value = event_emitter
    orchestrator = MagicMock(CLISessionOrchestrator)
    orchestrator.session_id = 'my-session-id'

    def fake_orchestrator_factory():
        return orchestrator

    register_session_id_event(
        session, orchestrator_factory=fake_orchestrator_factory
    )
    handler = event_emitter.register.call_args[0][1]
    handler()
    assert session.user_agent_extra == 'md/installer#source sid/my-session-id'
    event_emitter.unregister.assert_called_once_with(
        'before-create-client', handler
    )

def test_is_telemetry_disabled():
    assert is_telemetry_disabled({'AWS_CLI_SESSION_ID_DISABLED': 'true'})
    assert not is_telemetry_disabled({'AWS_CLI_SESSION_ID_DISABLED': 'false'})
    assert not is_telemetry_disabled({})


def test_register_session_id_event_skipped_when_telemetry_disabled():
    session = MagicMock(Session)
    event_emitter = MagicMock()
    session.get_component.return_value = event_emitter
    orchestrator_factory = MagicMock()

    with patch.dict(os.environ, {'AWS_CLI_SESSION_ID_DISABLED': 'true'}):
        register_session_id_event(
            session, orchestrator_factory=orchestrator_factory
        )

    # No handler is registered, so the session database is never opened.
    event_emitter.register.assert_not_called()
    orchestrator_factory.assert_not_called()


def test_disabled_telemetry_does_not_create_database(tmp_path):
    # The whole point of opting out is to avoid the database's file locking,
    # so ensure no database file is created at all.
    cache_dir = tmp_path / '.aws' / 'cli' / 'cache'
    with (
        patch.dict(os.environ, {'AWS_CLI_SESSION_ID_DISABLED': 'true'}),
        patch('awscli.telemetry._CACHE_DIR', cache_dir),
    ):
        driver = create_clidriver()
        driver.session.create_client('sts', region_name='us-east-1')

    assert not cache_dir.exists()


def test_register_session_id_event_catches_bare_exceptions():
    session = MagicMock(Session)
    session.user_agent_extra = ''
    event_emitter = MagicMock()
    session.get_component.return_value = event_emitter
    register_session_id_event(
        session, orchestrator_factory=MagicMock(side_effect=Exception)
    )
    handler = event_emitter.register.call_args[0][1]
    handler()
    assert session.user_agent_extra == ''


def test_user_agent_extra_contains_installer_component():
    # register_session_id_event depends on md/installer being present
    # in user_agent_extra to insert sid at the correct position. This
    # test ensures that invariant holds after driver creation.
    driver = create_clidriver()
    assert 'md/installer#' in driver.session.user_agent_extra



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput

from awscli.customizations.prompts import multiselect_choice

# ANSI escape sequences for simulating input
UP = '\x1b[A'
DOWN = '\x1b[B'
SPACE = ' '
ENTER = '\r'
CTRL_C = '\x03'


@pytest.fixture
def pipe_input():
    with create_pipe_input() as inp:
        yield inp


def _run(pipe_input, items, **kwargs):
    return multiselect_choice(
        '', items, pt_input=pipe_input, pt_output=DummyOutput(), **kwargs
    )


def test_enter_returns_all_preselected_by_default(pipe_input):
    pipe_input.send_text(ENTER)
    assert _run(pipe_input, ['a', 'b', 'c']) == ['a', 'b', 'c']


def test_space_deselects_current_item(pipe_input):
    pipe_input.send_text(SPACE + ENTER)
    assert _run(pipe_input, ['a', 'b', 'c']) == ['b', 'c']


def test_navigate_down_and_deselect(pipe_input):
    pipe_input.send_text(DOWN + SPACE + ENTER)
    assert _run(pipe_input, ['a', 'b', 'c']) == ['a', 'c']


def test_navigate_up_wraps_around(pipe_input):
    pipe_input.send_text(UP + SPACE + ENTER)
    assert _run(pipe_input, ['a', 'b', 'c']) == ['a', 'b']


def test_toggle_on_and_off(pipe_input):
    pipe_input.send_text(SPACE + SPACE + ENTER)
    assert _run(pipe_input, ['a', 'b', 'c']) == ['a', 'b', 'c']


def test_deselect_all(pipe_input):
    pipe_input.send_text(SPACE + DOWN + SPACE + DOWN + SPACE + ENTER)
    assert _run(pipe_input, ['a', 'b', 'c']) == []


def test_preselected_none_starts_all_selected(pipe_input):
    pipe_input.send_text(ENTER)
    assert _run(pipe_input, ['x', 'y'], preselected=None) == ['x', 'y']


def test_preselected_subset(pipe_input):
    pipe_input.send_text(ENTER)
    assert _run(pipe_input, ['a', 'b', 'c'], preselected={1}) == ['b']


def test_preselected_empty(pipe_input):
    pipe_input.send_text(ENTER)
    assert _run(pipe_input, ['a', 'b', 'c'], preselected=set()) == []


def test_display_format_does_not_affect_return_value(pipe_input):
    pipe_input.send_text(ENTER)
    assert _run(
        pipe_input, [1, 2, 3], display_format=lambda x: f'item-{x}'
    ) == [1, 2, 3]


def test_single_item(pipe_input):
    pipe_input.send_text(ENTER)
    assert _run(pipe_input, ['only']) == ['only']


def test_ctrl_c_raises_keyboard_interrupt(pipe_input):
    pipe_input.send_text(CTRL_C)
    with pytest.raises(KeyboardInterrupt):
        _run(pipe_input, ['a', 'b'])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import hashlib
import json
from io import StringIO
from unittest.mock import Mock, patch

import pytest

from awscli.customizations.agenttoolkit.agents import (
    SKILL_METADATA_FILENAME,
)
from awscli.customizations.agenttoolkit.update_skill import UpdateSkillCommand
from awscli.customizations.exceptions import ParamValidationError
from tests.unit.customizations.agenttoolkit.utils import (
    make_config,
    make_session,
    make_skill_zip,
)


def _install_skill_at_version(tmp_path, agent_dir, skill_name, version):
    skill_dir = tmp_path / agent_dir / 'skills' / skill_name
    skill_dir.mkdir(parents=True, exist_ok=True)
    (skill_dir / 'SKILL.md').write_text(f'name: {skill_name}\n')
    (skill_dir / SKILL_METADATA_FILENAME).write_text(
        json.dumps({'version': version})
    )


def _run_update(agent_configs, args, remote_version='v2', zip_bytes=None):
    if zip_bytes is not None:
        checksum = hashlib.sha256(zip_bytes).hexdigest()
    else:
        zip_bytes, checksum = make_skill_zip({'SKILL.md': 'test'})

    mock_client = Mock()
    mock_client.meta.endpoint_url = 'https://example.com'

    stream = StringIO()
    session = make_session()

    with (
        patch(
            'awscli.customizations.agenttoolkit.update_skill.resolve_latest_version',
            return_value=remote_version,
        ),
        patch(
            'awscli.customizations.agenttoolkit.update_skill.get_skill_download',
            return_value=(zip_bytes, checksum, remote_version),
        ),
        patch(
            'awscli.customizations.agenttoolkit.update_skill.create_client',
            return_value=mock_client,
        ),
    ):
        cmd = UpdateSkillCommand(
            session, stream=stream, agent_configs=agent_configs
        )
        rc = cmd(args=args, parsed_globals=Mock())
    return rc, stream.getvalue()


def test_update_skill_outdated(tmp_path):
    _install_skill_at_version(tmp_path, '.test-agent', 'aws-s3', 'v1')
    configs = [make_config(tmp_path)]
    rc, output = _run_update(
        configs, ['--skill-name', 'aws-s3'], remote_version='v2'
    )
    assert rc == 0
    assert 'Updated aws-s3 (v2)' in output
    marker = (
        tmp_path
        / '.test-agent'
        / 'skills'
        / 'aws-s3'
        / SKILL_METADATA_FILENAME
    )
    assert json.loads(marker.read_text()) == {'version': 'v2'}


def test_update_skill_already_up_to_date(tmp_path):
    _install_skill_at_version(tmp_path, '.test-agent', 'aws-s3', 'v1')
    configs = [make_config(tmp_path)]
    rc, output = _run_update(
        configs, ['--skill-name', 'aws-s3'], remote_version='v1'
    )
    assert rc == 0
    assert 'already up to date' in output


def test_update_skill_not_installed(tmp_path):
    (tmp_path / '.test-agent' / 'skills').mkdir(parents=True)
    configs = [make_config(tmp_path)]
    with pytest.raises(ParamValidationError, match='not installed'):
        _run_update(configs, ['--skill-name', 'aws-s3'])


def test_update_skill_only_outdated_agents_updated(tmp_path):
    _install_skill_at_version(tmp_path, '.agent-a', 'aws-s3', 'v1')
    _install_skill_at_version(tmp_path, '.agent-b', 'aws-s3', 'v2')
    configs = [
        make_config(
            tmp_path,
            id='agent-a',
            display_name='Agent A',
            detection_path=str(tmp_path / '.agent-a'),
        ),
        make_config(
            tmp_path,
            id='agent-b',
            display_name='Agent B',
            detection_path=str(tmp_path / '.agent-b'),
        ),
    ]
    rc, output = _run_update(
        configs, ['--skill-name', 'aws-s3'], remote_version='v2'
    )
    assert rc == 0
    assert 'Updated aws-s3 (v2) to Agent A' in output
    assert 'Agent B' not in output


def test_update_skill_missing_marker_skipped(tmp_path):
    skill_dir = tmp_path / '.test-agent' / 'skills' / 'aws-s3'
    skill_dir.mkdir(parents=True)
    (skill_dir / 'SKILL.md').write_text('test')
    configs = [make_config(tmp_path)]
    with pytest.raises(ParamValidationError, match='not installed'):
        _run_update(configs, ['--skill-name', 'aws-s3'], remote_version='v2')


def test_update_skill_through_symlinked_dir(tmp_path):
    # A user symlinks the per-skill dir into a shared location. update-skill
    # must follow the link and overwrite the target in place (shutil.rmtree
    # refuses to operate on the symlink itself), leaving the link intact.
    (tmp_path / '.test-agent' / 'skills').mkdir(parents=True)
    target = tmp_path / 'shared' / 'aws-s3'
    target.mkdir(parents=True)
    (target / 'SKILL.md').write_text('old')
    (target / 'old.md').write_text('removed in v2')
    (target / SKILL_METADATA_FILENAME).write_text(
        json.dumps({'version': 'v1'})
    )
    link = tmp_path / '.test-agent' / 'skills' / 'aws-s3'
    link.symlink_to(target, target_is_directory=True)

    configs = [make_config(tmp_path)]
    new_zip = make_skill_zip({'SKILL.md': 'new'})[0]
    rc, output = _run_update(
        configs,
        ['--skill-name', 'aws-s3'],
        remote_version='v2',
        zip_bytes=new_zip,
    )
    assert rc == 0
    assert 'Updated aws-s3 (v2)' in output
    assert link.is_symlink()
    assert (target / 'SKILL.md').read_text() == 'new'
    assert not (target / 'old.md').exists()
    assert json.loads((target / SKILL_METADATA_FILENAME).read_text()) == {
        'version': 'v2'
    }


def test_update_skill_removes_orphaned_files(tmp_path):
    skill_dir = tmp_path / '.test-agent' / 'skills' / 'aws-s3'
    skill_dir.mkdir(parents=True)
    (skill_dir / 'SKILL.md').write_text('old')
    (skill_dir / 'old.md').write_text('removed in v2')
    (skill_dir / SKILL_METADATA_FILENAME).write_text(
        json.dumps({'version': 'v1'})
    )
    configs = [make_config(tmp_path)]
    new_zip = make_skill_zip({'SKILL.md': 'new'})[0]
    rc, _ = _run_update(
        configs,
        ['--skill-name', 'aws-s3'],
        remote_version='v2',
        zip_bytes=new_zip,
    )
    assert rc == 0
    assert (skill_dir / 'SKILL.md').read_text() == 'new'
    assert not (skill_dir / 'old.md').exists()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import hashlib
import io
import json
import zipfile
from unittest.mock import Mock

from awscli.customizations.agenttoolkit.agents import (
    SKILL_METADATA_FILENAME,
    AgentConfig,
)


def make_parsed_globals(output='json'):
    pg = Mock()
    pg.output = output
    pg.query = None
    pg.color = 'off'
    pg.no_paginate = False
    return pg


def make_session():
    session = Mock()
    session.user_agent_extra = ''
    session.get_config_variable.return_value = 'json'
    session.emit_first_non_none_response.return_value = None
    return session


def make_skill_zip(files=None):
    """Create a skill zip and returns (bytes, checksum)."""
    if files is None:
        files = {'SKILL.md': '# Test Skill\n'}
    buf = io.BytesIO()
    with zipfile.ZipFile(buf, 'w') as zf:
        for name, content in files.items():
            zf.writestr(name, content)
    zip_bytes = buf.getvalue()
    checksum = hashlib.sha256(zip_bytes).hexdigest()
    return zip_bytes, checksum


def make_skill(tmp_path, agent_name, skill_name, with_marker=True):
    skill_dir = tmp_path / agent_name / 'skills' / skill_name
    skill_dir.mkdir(parents=True)
    (skill_dir / 'SKILL.md').write_text('test')
    if with_marker:
        (skill_dir / SKILL_METADATA_FILENAME).write_text(
            json.dumps({'version': 'v1'})
        )


def make_config(tmp_path, **overrides):
    defaults = {
        'id': 'test-agent',
        'display_name': 'Test Agent',
        'detection_path': str(tmp_path / '.test-agent'),
    }
    defaults.update(overrides)
    return AgentConfig(**defaults)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
import subprocess
from argparse import Namespace
from unittest import mock

import pytest

from awscli.customizations import update as update_module
from awscli.customizations.update import (
    UnixUpdateCommand,
    UpdateError,
    WindowsUpdateCommand,
)
from tests.markers import skip_if_windows

USER_INSTALL = {
    'install_dir': '/home/user/.local/share/aws-cli',
    'bin_dir': '/home/user/.local/bin',
    'script_install': {'system': False},
}
SYSTEM_INSTALL = {
    'install_dir': '/usr/local/aws-cli',
    'bin_dir': '/usr/local/bin',
    'script_install': {'system': True},
}


def global_args(color='auto'):
    return Namespace(color=color)


def mock_session():
    session = mock.Mock()
    session.user_agent_extra = 'aws-cli'
    return session


class TestUnixUpdateCommand:
    def _command(
        self,
        install,
        source='exe',
        elevated=True,
        runner=None,
        downloader=None,
    ):
        return UnixUpdateCommand(
            mock_session(),
            source=source,
            install_metadata=install,
            downloader=downloader or mock.Mock(),
            is_elevated=elevated,
            runner=runner or mock.Mock(),
        )

    def _run(self, install, color='auto', **kwargs):
        runner = mock.Mock()
        command = self._command(install, runner=runner, **kwargs)
        command([], global_args(color=color))
        return runner.call_args.args

    @pytest.mark.parametrize('source', ['exe', 'script-exe', 'update-exe'])
    def test_supported_distribution_source_runs_installer(self, source):
        runner = mock.Mock()
        command = self._command(USER_INSTALL, source=source, runner=runner)

        assert command([], global_args()) == 0
        runner.assert_called_once()

    @pytest.mark.parametrize('source', ['source', 'other', 'pip', ''])
    def test_unsupported_distribution_source_raises(self, source):
        runner = mock.Mock()
        command = self._command(USER_INSTALL, source=source, runner=runner)

        with pytest.raises(UpdateError, match='Detected distribution source'):
            command([], global_args())
        runner.assert_not_called()

    def test_missing_install_dir_raises(self):
        runner = mock.Mock()
        command = self._command({}, runner=runner)

        with pytest.raises(
            UpdateError, match='Install-time metadata could not be found'
        ):
            command([], global_args())
        runner.assert_not_called()

    def test_user_install_runs_bash_script_without_system(self):
        cmd, _ = self._run(USER_INSTALL)

        assert cmd[0] == 'bash'
        assert cmd[1].endswith('install.sh')
        assert len(cmd) == 2

    def test_downloads_install_script_to_path_that_is_run(self):
        downloader = mock.Mock()
        runner = mock.Mock()
        command = self._command(
            USER_INSTALL, downloader=downloader, runner=runner
        )

        command([], global_args())

        url, dest = downloader.call_args.args
        assert url == UnixUpdateCommand.SCRIPT_URL
        assert runner.call_args.args[0] == ['bash', dest]

    def test_color_off_sets_no_color_env(self):
        _, env = self._run(USER_INSTALL, color='off')

        assert env['NO_COLOR'] == '1'

    @pytest.mark.parametrize('color', ['auto', 'on'])
    def test_color_not_off_omits_no_color_env(self, color):
        _, env = self._run(USER_INSTALL, color=color)

        assert 'NO_COLOR' not in env

    def test_user_install_sets_xdg_from_install_dir(self):
        _, env = self._run(USER_INSTALL)

        assert env['XDG_DATA_HOME'] == '/home/user/.local/share'
        assert env['XDG_BIN_HOME'] == '/home/user/.local/bin'
        assert env['AWS_CLI_DISTRIBUTION_SOURCE_OVERRIDE'] == 'update-exe'
        assert 'AWS_CLI_NO_BIN_DIR' not in env

    def test_missing_bin_dir_uses_throwaway_dir_and_flag(self):
        install = {
            'install_dir': '/home/user/.local/share/aws-cli',
            'script_install': {'system': False},
        }

        _, env = self._run(install)

        assert env['AWS_CLI_NO_BIN_DIR'] == '1'
        assert env['XDG_BIN_HOME'].endswith('bin')
        assert '.local/bin' not in env['XDG_BIN_HOME']

    def test_system_install_passes_system_flag(self):
        cmd, env = self._run(SYSTEM_INSTALL, elevated=True)

        assert cmd[0] == 'bash'
        assert cmd[1].endswith('install.sh')
        assert cmd[2] == '--system'
        assert len(cmd) == 3
        assert 'XDG_DATA_HOME' not in env
        assert 'XDG_BIN_HOME' not in env

    def test_system_install_requires_elevation(self):
        runner = mock.Mock()
        command = self._command(SYSTEM_INSTALL, elevated=False, runner=runner)

        with pytest.raises(UpdateError, match='requires root'):
            command([], global_args())
        runner.assert_not_called()

    def test_install_script_failure_raises_update_error(self):
        runner = mock.Mock(
            side_effect=subprocess.CalledProcessError(8, ['bash'])
        )
        command = self._command(USER_INSTALL, runner=runner)

        with pytest.raises(UpdateError, match='exit code 8'):
            command([], global_args())

    @skip_if_windows
    def test_system_detected_from_binary_path_when_no_metadata(
        self, monkeypatch
    ):
        monkeypatch.setattr(
            update_module.sys,
            'executable',
            '/usr/local/aws-cli/v2/current/bin/aws',
        )

        cmd, _ = self._run({'install_dir': '/usr/local/aws-cli'})

        assert '--system' in cmd

    @skip_if_windows
    def test_user_detected_from_binary_path_when_no_metadata(
        self, monkeypatch
    ):
        monkeypatch.setattr(
            update_module.sys,
            'executable',
            '/home/user/.local/share/aws-cli/v2/current/bin/aws',
        )
        install = {
            'install_dir': '/home/user/.local/share/aws-cli',
            'bin_dir': '/home/user/.local/bin',
        }

        cmd, _ = self._run(install)

        assert '--system' not in cmd

    @skip_if_windows
    def test_resolves_symlink_before_checking_install_dir(
        self, monkeypatch, tmp_path
    ):
        root = tmp_path.resolve()
        install_dir = root / 'aws-cli'
        real_bin = install_dir / 'v2' / 'current' / 'bin' / 'aws'
        real_bin.parent.mkdir(parents=True)
        real_bin.touch()
        link = root / 'bin' / 'aws'
        link.parent.mkdir()
        link.symlink_to(real_bin)

        monkeypatch.setattr(
            UnixUpdateCommand, 'SYSTEM_INSTALL_DIR', str(install_dir)
        )
        monkeypatch.setattr(update_module.sys, 'executable', str(link))

        cmd, _ = self._run({'install_dir': str(install_dir)})

        assert '--system' in cmd


class TestWindowsUpdateCommand:
    def _command(self, install, elevated=True, runner=None, downloader=None):
        return WindowsUpdateCommand(
            mock_session(),
            source='exe',
            install_metadata=install,
            downloader=downloader or mock.Mock(),
            is_elevated=elevated,
            runner=runner or mock.Mock(),
            powershell_path='powershell.exe',
        )

    def _run(self, install, color='auto', **kwargs):
        runner = mock.Mock()
        command = self._command(install, runner=runner, **kwargs)
        command([], global_args(color=color))
        (cmd,) = runner.call_args.args
        with open(cmd[-1]) as f:
            return cmd, f.read()

    def test_spawns_detached_cmd_wrapper(self):
        cmd, _ = self._run(USER_INSTALL)

        assert cmd[0] == 'cmd'
        assert cmd[1] == '/c'
        assert cmd[2].endswith('.cmd')
        assert len(cmd) == 3

    def test_downloads_install_script_referenced_by_wrapper(self):
        downloader = mock.Mock()
        _, wrapper = self._run(USER_INSTALL, downloader=downloader)

        url, dest = downloader.call_args.args
        assert url == WindowsUpdateCommand.SCRIPT_URL
        assert f'-File "{dest}"' in wrapper

    def test_wrapper_runs_powershell_without_system(self):
        _, wrapper = self._run(USER_INSTALL)

        assert 'set AWS_CLI_DISTRIBUTION_SOURCE_OVERRIDE=update-exe' in wrapper
        assert '"powershell.exe" -NoProfile -File "' in wrapper
        assert '-System' not in wrapper
        assert 'set NO_COLOR=1' not in wrapper

    def test_wrapper_passes_system_flag(self):
        _, wrapper = self._run(SYSTEM_INSTALL, elevated=True)

        assert wrapper.rstrip().endswith('-System')

    def test_wrapper_sets_no_color_when_color_off(self):
        _, wrapper = self._run(USER_INSTALL, color='off')

        assert 'set NO_COLOR=1' in wrapper

    def test_system_install_requires_elevation(self):
        runner = mock.Mock()
        command = self._command(SYSTEM_INSTALL, elevated=False, runner=runner)

        with pytest.raises(UpdateError, match='elevated shell'):
            command([], global_args())
        runner.assert_not_called()

    def test_system_detected_from_program_files_when_no_metadata(
        self, monkeypatch
    ):
        program_files = 'C:\\Program Files'
        monkeypatch.setattr(
            update_module.os, 'environ', {'ProgramW6432': program_files}
        )
        running_exe = update_module.os.path.join(
            program_files, 'Amazon', 'AWSCLIV2', 'aws.exe'
        )
        monkeypatch.setattr(update_module, 'ctypes', _fake_ctypes(running_exe))

        _, wrapper = self._run(
            {'install_dir': 'C:\\Program Files\\Amazon\\AWSCLIV2'}
        )

        assert wrapper.rstrip().endswith('-System')

    def test_user_detected_from_program_files_when_no_metadata(
        self, monkeypatch
    ):
        monkeypatch.setattr(
            update_module.os, 'environ', {'ProgramW6432': 'C:\\Program Files'}
        )
        running_exe = update_module.os.path.join(
            'C:\\Users\\me\\AppData\\Local\\Programs',
            'Amazon',
            'AWSCLIV2',
            'aws.exe',
        )
        monkeypatch.setattr(update_module, 'ctypes', _fake_ctypes(running_exe))

        _, wrapper = self._run({'install_dir': 'C:\\Users\\me\\AWSCLIV2'})

        assert '-System' not in wrapper


def _response(status_code, content=b''):
    return mock.Mock(status_code=status_code, content=content)


class TestDownloadWithRetry:
    def test_sends_get_request_for_url(self, tmp_path):
        session = mock.Mock()
        session.send.return_value = _response(200, b'data')

        update_module.download_with_retry(
            'https://foo.bar/install.sh',
            str(tmp_path / 'install.sh'),
            session=session,
        )

        request = session.send.call_args.args[0]
        assert request.method == 'GET'
        assert request.url == 'https://foo.bar/install.sh'

    def test_writes_response_content_to_dest(self, tmp_path):
        session = mock.Mock()
        session.send.return_value = _response(200, b'#!/bin/sh\n')
        dest = tmp_path / 'install.sh'

        update_module.download_with_retry(
            'https://foo.bar/install.sh', str(dest), session=session
        )

        assert dest.read_bytes() == b'#!/bin/sh\n'

    def test_retries_once_then_succeeds(self, tmp_path):
        session = mock.Mock()
        session.send.side_effect = [
            ConnectionError('boom'),
            _response(200, b'ok'),
        ]
        dest = tmp_path / 'install.sh'

        update_module.download_with_retry(
            'https://foo.bar/install.sh', str(dest), session=session
        )

        assert dest.read_bytes() == b'ok'
        assert session.send.call_count == 2

    def test_non_200_status_raises(self, tmp_path):
        session = mock.Mock()
        session.send.return_value = _response(404)

        with pytest.raises(UpdateError, match='unexpected HTTP status 404'):
            update_module.download_with_retry(
                'https://foo.bar/install.sh',
                str(tmp_path / 'install.sh'),
                session=session,
            )

    def test_raises_after_retries_exhausted(self, tmp_path):
        session = mock.Mock()
        session.send.side_effect = ConnectionError('boom')

        with pytest.raises(UpdateError, match='failed to download'):
            update_module.download_with_retry(
                'https://foo.bar/install.sh',
                str(tmp_path / 'install.sh'),
                retries=1,
                session=session,
            )
        assert session.send.call_count == 2

    def test_honors_retry_count(self, tmp_path):
        session = mock.Mock()
        session.send.side_effect = ConnectionError('boom')

        with pytest.raises(UpdateError):
            update_module.download_with_retry(
                'https://foo.bar/install.sh',
                str(tmp_path / 'install.sh'),
                retries=3,
                session=session,
            )
        assert session.send.call_count == 4


def _fake_ctypes(module_filename):
    buf = mock.MagicMock()
    buf.value = module_filename
    buf.__len__.return_value = 32768
    fake = mock.Mock()
    fake.create_unicode_buffer.return_value = buf
    return fake



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io

import pytest
from botocore.credentials import Credentials, ReadOnlyCredentials
from botocore.exceptions import (
    ClientError,
    InvalidConfigError,
    NoCredentialsError,
)
from botocore.session import Session
from s3transfer.constants import GB
from s3transfer.exceptions import TransferNotDoneError
from s3transfer.utils import CallArgs

from tests import HAS_CRT, FileCreator, mock, requires_crt, unittest

if HAS_CRT:
    import awscrt.auth
    import awscrt.s3
    import s3transfer.crt


requires_crt_pytest = pytest.mark.skipif(
    not HAS_CRT, reason="Test requires awscrt to be installed."
)


@pytest.fixture
def mock_crt_process_lock(monkeypatch):
    # The process lock is cached at the module layer whenever the
    # cross process lock is successfully acquired. This patch ensures that
    # test cases will start off with no previously cached process lock and
    # if a cross process is instantiated/acquired it will be the mock that
    # can be used for controlling lock behavior.
    monkeypatch.setattr('s3transfer.crt.CRT_S3_PROCESS_LOCK', None)
    with mock.patch('awscrt.s3.CrossProcessLock', spec=True) as mock_lock:
        yield mock_lock


@pytest.fixture
def mock_s3_crt_client():
    with mock.patch('s3transfer.crt.S3Client', spec=True) as mock_client:
        yield mock_client


@pytest.fixture
def mock_get_recommended_throughput_target_gbps():
    with mock.patch(
        's3transfer.crt.get_recommended_throughput_target_gbps'
    ) as mock_get_target_gbps:
        yield mock_get_target_gbps


class CustomFutureException(Exception):
    pass


@requires_crt_pytest
class TestCRTProcessLock:
    def test_acquire_crt_s3_process_lock(self, mock_crt_process_lock):
        lock = s3transfer.crt.acquire_crt_s3_process_lock('app-name')
        assert lock is s3transfer.crt.CRT_S3_PROCESS_LOCK
        assert lock is mock_crt_process_lock.return_value
        mock_crt_process_lock.assert_called_once_with('app-name')
        mock_crt_process_lock.return_value.acquire.assert_called_once_with()

    def test_unable_to_acquire_lock_returns_none(self, mock_crt_process_lock):
        mock_crt_process_lock.return_value.acquire.side_effect = RuntimeError
        assert s3transfer.crt.acquire_crt_s3_process_lock('app-name') is None
        assert s3transfer.crt.CRT_S3_PROCESS_LOCK is None
        mock_crt_process_lock.assert_called_once_with('app-name')
        mock_crt_process_lock.return_value.acquire.assert_called_once_with()

    def test_multiple_acquires_return_same_lock(self, mock_crt_process_lock):
        lock = s3transfer.crt.acquire_crt_s3_process_lock('app-name')
        assert s3transfer.crt.acquire_crt_s3_process_lock('app-name') is lock
        assert lock is s3transfer.crt.CRT_S3_PROCESS_LOCK

        # The process lock should have only been instantiated and acquired once
        mock_crt_process_lock.assert_called_once_with('app-name')
        mock_crt_process_lock.return_value.acquire.assert_called_once_with()


@requires_crt()
class TestBotocoreCRTRequestSerializer(unittest.TestCase):
    def setUp(self):
        self.region = 'us-west-2'
        self.session = Session()
        self.session.set_config_variable('region', self.region)
        self.request_serializer = s3transfer.crt.BotocoreCRTRequestSerializer(
            self.session
        )
        self.bucket = "test_bucket"
        self.key = "test_key"
        self.files = FileCreator()
        self.filename = self.files.create_file('myfile', 'my content')
        self.expected_path = "/" + self.bucket + "/" + self.key
        self.expected_host = f"s3.{self.region}.amazonaws.com"

    def tearDown(self):
        self.files.remove_all()

    def test_upload_request(self):
        callargs = CallArgs(
            bucket=self.bucket,
            key=self.key,
            fileobj=self.filename,
            extra_args={},
            subscribers=[],
        )
        coordinator = s3transfer.crt.CRTTransferCoordinator()
        future = s3transfer.crt.CRTTransferFuture(
            s3transfer.crt.CRTTransferMeta(call_args=callargs), coordinator
        )
        crt_request = self.request_serializer.serialize_http_request(
            "put_object", future
        )
        self.assertEqual("PUT", crt_request.method)
        self.assertEqual(self.expected_path, crt_request.path)
        self.assertEqual(self.expected_host, crt_request.headers.get("host"))
        self.assertIsNone(crt_request.headers.get("Authorization"))

    def test_download_request(self):
        callargs = CallArgs(
            bucket=self.bucket,
            key=self.key,
            fileobj=self.filename,
            extra_args={},
            subscribers=[],
        )
        coordinator = s3transfer.crt.CRTTransferCoordinator()
        future = s3transfer.crt.CRTTransferFuture(
            s3transfer.crt.CRTTransferMeta(call_args=callargs), coordinator
        )
        crt_request = self.request_serializer.serialize_http_request(
            "get_object", future
        )
        self.assertEqual("GET", crt_request.method)
        self.assertEqual(self.expected_path, crt_request.path)
        self.assertEqual(self.expected_host, crt_request.headers.get("host"))
        self.assertIsNone(crt_request.headers.get("Authorization"))

    def test_delete_request(self):
        callargs = CallArgs(
            bucket=self.bucket, key=self.key, extra_args={}, subscribers=[]
        )
        coordinator = s3transfer.crt.CRTTransferCoordinator()
        future = s3transfer.crt.CRTTransferFuture(
            s3transfer.crt.CRTTransferMeta(call_args=callargs), coordinator
        )
        crt_request = self.request_serializer.serialize_http_request(
            "delete_object", future
        )
        self.assertEqual("DELETE", crt_request.method)
        self.assertEqual(self.expected_path, crt_request.path)
        self.assertEqual(self.expected_host, crt_request.headers.get("host"))
        self.assertIsNone(crt_request.headers.get("Authorization"))

    def _create_crt_response_error(
        self, status_code, body, operation_name=None
    ):
        return awscrt.s3.S3ResponseError(
            code=14343,
            name='AWS_ERROR_S3_INVALID_RESPONSE_STATUS',
            message='Invalid response status from request',
            status_code=status_code,
            headers=[
                ('x-amz-request-id', 'QSJHJJZR2EDYD4GQ'),
                (
                    'x-amz-id-2',
                    'xDbgdKdvYZTjgpOTzm7yNP2JPrOQl+eaQvUkFdOjdJoWkIC643fgHxdsHpUKvVAfjKf5F6otEYA=',
                ),
                ('Content-Type', 'application/xml'),
                ('Transfer-Encoding', 'chunked'),
                ('Date', 'Fri, 10 Nov 2023 23:22:47 GMT'),
                ('Server', 'AmazonS3'),
            ],
            body=body,
            operation_name=operation_name,
        )

    def test_translate_get_object_404(self):
        body = (
            b'<?xml version="1.0" encoding="UTF-8"?>\n<Error>'
            b'<Code>NoSuchKey</Code>'
            b'<Message>The specified key does not exist.</Message>'
            b'<Key>obviously-no-such-key.txt</Key>'
            b'<RequestId>SBJ7ZQY03N1WDW9T</RequestId>'
            b'<HostId>SomeHostId</HostId></Error>'
        )
        crt_exc = self._create_crt_response_error(404, body, 'GetObject')
        boto_err = self.request_serializer.translate_crt_exception(crt_exc)
        self.assertIsInstance(
            boto_err, self.session.create_client('s3').exceptions.NoSuchKey
        )

    def test_translate_head_object_404(self):
        # There's no body in a HEAD response, so we can't map it to a modeled S3 exception.
        # But it should still map to a botocore ClientError
        body = None
        crt_exc = self._create_crt_response_error(
            404, body, operation_name='HeadObject'
        )
        boto_err = self.request_serializer.translate_crt_exception(crt_exc)
        self.assertIsInstance(boto_err, ClientError)

    def test_translate_unknown_operation_404(self):
        body = None
        crt_exc = self._create_crt_response_error(404, body)
        boto_err = self.request_serializer.translate_crt_exception(crt_exc)
        self.assertIsInstance(boto_err, ClientError)


@requires_crt_pytest
class TestBotocoreCRTCredentialsWrapper:
    @pytest.fixture
    def botocore_credentials(self):
        return Credentials(
            access_key='access_key', secret_key='secret_key', token='token'
        )

    def assert_crt_credentials(
        self,
        crt_credentials,
        expected_access_key='access_key',
        expected_secret_key='secret_key',
        expected_token='token',
    ):
        assert crt_credentials.access_key_id == expected_access_key
        assert crt_credentials.secret_access_key == expected_secret_key
        assert crt_credentials.session_token == expected_token

    def test_fetch_crt_credentials_successfully(self, botocore_credentials):
        wrapper = s3transfer.crt.BotocoreCRTCredentialsWrapper(
            botocore_credentials
        )
        crt_credentials = wrapper()
        self.assert_crt_credentials(crt_credentials)

    def test_wrapper_does_not_cache_frozen_credentials(self):
        mock_credentials = mock.Mock(Credentials)
        mock_credentials.get_frozen_credentials.side_effect = [
            ReadOnlyCredentials('access_key_1', 'secret_key_1', 'token_1'),
            ReadOnlyCredentials('access_key_2', 'secret_key_2', 'token_2'),
        ]
        wrapper = s3transfer.crt.BotocoreCRTCredentialsWrapper(
            mock_credentials
        )

        crt_credentials_1 = wrapper()
        self.assert_crt_credentials(
            crt_credentials_1,
            expected_access_key='access_key_1',
            expected_secret_key='secret_key_1',
            expected_token='token_1',
        )

        crt_credentials_2 = wrapper()
        self.assert_crt_credentials(
            crt_credentials_2,
            expected_access_key='access_key_2',
            expected_secret_key='secret_key_2',
            expected_token='token_2',
        )

        assert mock_credentials.get_frozen_credentials.call_count == 2

    def test_raises_error_when_resolved_credentials_is_none(self):
        wrapper = s3transfer.crt.BotocoreCRTCredentialsWrapper(None)
        with pytest.raises(NoCredentialsError):
            wrapper()

    def test_to_crt_credentials_provider(self, botocore_credentials):
        wrapper = s3transfer.crt.BotocoreCRTCredentialsWrapper(
            botocore_credentials
        )
        crt_credentials_provider = wrapper.to_crt_credentials_provider()
        assert isinstance(
            crt_credentials_provider, awscrt.auth.AwsCredentialsProvider
        )
        get_credentials_future = crt_credentials_provider.get_credentials()
        crt_credentials = get_credentials_future.result()
        self.assert_crt_credentials(crt_credentials)


@requires_crt()
class TestCRTTransferFuture(unittest.TestCase):
    def setUp(self):
        self.mock_s3_request = mock.Mock(awscrt.s3.S3RequestType)
        self.mock_crt_future = mock.Mock(awscrt.s3.Future)
        self.mock_s3_request.finished_future = self.mock_crt_future
        self.coordinator = s3transfer.crt.CRTTransferCoordinator()
        self.coordinator.set_s3_request(self.mock_s3_request)
        self.future = s3transfer.crt.CRTTransferFuture(
            coordinator=self.coordinator
        )

    def test_set_exception(self):
        self.future.set_exception(CustomFutureException())
        with self.assertRaises(CustomFutureException):
            self.future.result()

    def test_set_exception_raises_error_when_not_done(self):
        self.mock_crt_future.done.return_value = False
        with self.assertRaises(TransferNotDoneError):
            self.future.set_exception(CustomFutureException())

    def test_set_exception_can_override_previous_exception(self):
        self.future.set_exception(Exception())
        self.future.set_exception(CustomFutureException())
        with self.assertRaises(CustomFutureException):
            self.future.result()


@requires_crt()
class TestOnBodyFileObjWriter(unittest.TestCase):
    def test_call(self):
        fileobj = io.BytesIO()
        writer = s3transfer.crt.OnBodyFileObjWriter(fileobj)
        writer(chunk=b'content')
        self.assertEqual(fileobj.getvalue(), b'content')


@requires_crt_pytest
class TestCreateS3CRTClient:
    @pytest.mark.parametrize(
        'provided_bytes_per_sec,recommended_gbps,expected_gbps',
        [
            (None, 100.0, 100.0),
            (None, None, 10.0),
            # NOTE: create_s3_crt_client() accepts target throughput as bytes
            # per second and it is converted to gigabits per second for the
            # CRT client instantiation.
            (1_000_000_000, None, 8.0),
            (1_000_000_000, 100.0, 8.0),
        ],
    )
    def test_target_throughput(
        self,
        provided_bytes_per_sec,
        recommended_gbps,
        expected_gbps,
        mock_s3_crt_client,
        mock_get_recommended_throughput_target_gbps,
    ):
        mock_get_recommended_throughput_target_gbps.return_value = (
            recommended_gbps
        )
        s3transfer.crt.create_s3_crt_client(
            'us-west-2',
            target_throughput=provided_bytes_per_sec,
        )
        assert (
            mock_s3_crt_client.call_args[1]['throughput_target_gbps']
            == expected_gbps
        )

    def test_always_enables_s3express(self, mock_s3_crt_client):
        s3transfer.crt.create_s3_crt_client('us-west-2')
        assert mock_s3_crt_client.call_args[1]['enable_s3express'] is True

    def test_empty_verify_value_raises(self, mock_s3_crt_client):
        with pytest.raises(InvalidConfigError):
            s3transfer.crt.create_s3_crt_client('us-west-2', verify='')

    def test_whitespace_verify_value_raises(self, mock_s3_crt_client):
        with pytest.raises(InvalidConfigError):
            s3transfer.crt.create_s3_crt_client('us-west-2', verify='   ')

    def test_verify_false_disables_verification(self, mock_s3_crt_client):
        with (
            mock.patch('s3transfer.crt.TlsContextOptions') as mock_tls_options,
            mock.patch('s3transfer.crt.ClientTlsContext'),
        ):
            s3transfer.crt.create_s3_crt_client('us-west-2', verify=False)
        assert mock_tls_options.return_value.verify_peer is False

    @pytest.mark.parametrize(
        'fio_options,should_stream,disk_throughput,direct_io',
        [
            ({'should_stream': True}, True, 0.0, False),
            ({'disk_throughput_gbps': 8}, False, 8, False),
            ({'direct_io': True}, False, 0.0, True),
            (
                {'should_stream': True, 'disk_throughput_gbps': 8},
                True,
                8,
                False,
            ),
            ({'should_stream': True, 'direct_io': True}, True, 0.0, True),
            ({'disk_throughput_gbps': 8, 'direct_io': True}, False, 8, True),
            (
                {
                    'should_stream': True,
                    'disk_throughput_gbps': 8,
                    'direct_io': True,
                },
                True,
                8,
                True,
            ),
        ],
    )
    def test_fio_options(
        self,
        fio_options,
        should_stream,
        disk_throughput,
        direct_io,
        mock_s3_crt_client,
    ):
        params = {'fio_options': fio_options}
        s3transfer.crt.create_s3_crt_client(
            'us-west-2',
            **params,
        )
        assert (
            mock_s3_crt_client.call_args[1]['fio_options'].should_stream
            is should_stream
        )
        assert (
            mock_s3_crt_client.call_args[1]['fio_options'].disk_throughput_gbps
            == disk_throughput
        )
        assert (
            mock_s3_crt_client.call_args[1]['fio_options'].direct_io
            is direct_io
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import importlib

from awscli.commands import CLICommand


class LazyCommand(CLICommand):
    """A command-table entry that defers importing its real implementation.

    Sits in the command table like any other CLICommand, but only imports
    the actual module (and creates the real command object) when the command
    is invoked or its help is accessed.
    """

    def __init__(self, name, session, module_path, class_name):
        self._name = name
        self._session = session
        self._module_path = module_path
        self._class_name = class_name
        self._real = None
        self._lineage = [self]

    def _resolve(self):
        if self._real is None:
            mod = importlib.import_module(self._module_path)
            cls = getattr(mod, self._class_name)
            self._real = cls(self._session)
            self._real.lineage = self._lineage
        return self._real

    def __call__(self, args, parsed_globals):
        return self._resolve()(args, parsed_globals)

    def create_help_command(self):
        return self._resolve().create_help_command()

    @property
    def arg_table(self):
        return self._resolve().arg_table

    @property
    def subcommand_table(self):
        return self._resolve().subcommand_table

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def lineage(self):
        return self._lineage

    @lineage.setter
    def lineage(self, value):
        self._lineage = value
        if self._real is not None:
            self._real.lineage = value



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from unittest.mock import MagicMock, patch

import pytest

from awscli.handlers_registry import CommandTableOp
from awscli.lazy import LazyCommand
from awscli.lazy_emitter import LazyInitEmitter


@pytest.fixture
def mock_module():
    """Create a mock module with a callable init function."""
    module = MagicMock()
    module.my_init = MagicMock()
    return module


class TestLazyInitEmitterPrefixMatching:
    def test_bare_prefix_entry_initialized_on_dotted_emit(self, mock_module):
        # Entry registered against 'building-command-table' (bare prefix)
        # should be initialized when 'building-command-table.main' is emitted.
        registry = {
            'building-command-table': [
                ('test.module', 'my_init'),
            ],
        }
        emitter = LazyInitEmitter()
        emitter.load_registry(registry)

        with patch('importlib.import_module', return_value=mock_module):
            emitter.emit('building-command-table.main', command_table={})

        assert emitter.initialized_count == 1
        mock_module.my_init.assert_called_once()

    def test_exact_match_entry_initialized(self, mock_module):
        registry = {
            'building-command-table.main': [
                ('test.module', 'my_init'),
            ],
        }
        emitter = LazyInitEmitter()
        emitter.load_registry(registry)

        with patch('importlib.import_module', return_value=mock_module):
            emitter.emit('building-command-table.main', command_table={})

        assert emitter.initialized_count == 1
        mock_module.my_init.assert_called_once()

    def test_unrelated_entry_not_initialized(self, mock_module):
        # Entry registered against 'building-command-table.ecs' should NOT
        # be initialized when 'building-command-table.main' is emitted.
        registry = {
            'building-command-table.ecs': [
                ('test.module', 'my_init'),
            ],
        }
        emitter = LazyInitEmitter()
        emitter.load_registry(registry)

        with patch('importlib.import_module', return_value=mock_module):
            emitter.emit('building-command-table.main', command_table={})

        assert emitter.initialized_count == 0
        mock_module.my_init.assert_not_called()

    def test_multiple_prefix_levels_all_initialized(self, mock_module):
        # Both 'building-command-table' and 'building-command-table.main'
        # entries should be initialized when 'building-command-table.main'
        # is emitted.
        mock_module.other_init = MagicMock()
        registry = {
            'building-command-table': [
                ('test.module', 'my_init'),
            ],
            'building-command-table.main': [
                ('test.module', 'other_init'),
            ],
        }
        emitter = LazyInitEmitter()
        emitter.load_registry(registry)

        with patch('importlib.import_module', return_value=mock_module):
            emitter.emit('building-command-table.main', command_table={})

        assert emitter.initialized_count == 2
        mock_module.my_init.assert_called_once()
        mock_module.other_init.assert_called_once()

    def test_entry_initialized_only_once(self, mock_module):
        registry = {
            'building-command-table': [
                ('test.module', 'my_init'),
            ],
        }
        emitter = LazyInitEmitter()
        emitter.load_registry(registry)

        with patch('importlib.import_module', return_value=mock_module):
            emitter.emit('building-command-table.main', command_table={})
            emitter.emit('building-command-table.ecs', command_table={})

        assert emitter.initialized_count == 1
        mock_module.my_init.assert_called_once()


class TestMainCommandTableOps:
    def test_covered_plugins_not_imported(self, mock_module):
        registry = {
            'building-command-table.main': [
                ('heavy.module', 'register_add_cmd'),
            ],
        }
        main_ops = [
            (CommandTableOp.ADD, 'my-cmd', 'heavy.module', 'MyCommand'),
        ]
        emitter = LazyInitEmitter(main_command_table_ops=main_ops)
        emitter.load_registry(registry)

        command_table = {'existing': MagicMock(name='existing')}
        session = MagicMock()

        with patch('importlib.import_module', return_value=mock_module) as imp:
            emitter.emit(
                'building-command-table.main',
                command_table=command_table,
                session=session,
            )

        # The heavy module should NOT have been imported via _ensure_initialized
        imp.assert_not_called()
        # But the entry should be marked as initialized
        assert emitter.initialized_count == 1

    def test_rename_op_applied_to_command_table(self):
        registry = {
            'building-command-table.main': [
                ('rename.module', 'register_rename'),
            ],
        }
        main_ops = [
            (CommandTableOp.RENAME, 'old-name', 'new-name'),
        ]
        emitter = LazyInitEmitter(main_command_table_ops=main_ops)
        emitter.load_registry(registry)

        cmd = MagicMock()
        cmd.name = 'old-name'
        command_table = {'old-name': cmd}
        session = MagicMock()

        emitter.emit(
            'building-command-table.main',
            command_table=command_table,
            session=session,
        )

        assert 'old-name' not in command_table
        assert 'new-name' in command_table
        assert command_table['new-name'] is cmd
        assert cmd.name == 'new-name'

    def test_add_op_creates_lazy_command(self):
        registry = {
            'building-command-table.main': [
                ('add.module', 'register_add'),
            ],
        }
        main_ops = [
            (CommandTableOp.ADD, 'my-cmd', 'add.module.impl', 'MyCommand'),
        ]
        emitter = LazyInitEmitter(main_command_table_ops=main_ops)
        emitter.load_registry(registry)

        command_table = {}
        session = MagicMock()

        emitter.emit(
            'building-command-table.main',
            command_table=command_table,
            session=session,
        )

        assert 'my-cmd' in command_table
        assert isinstance(command_table['my-cmd'], LazyCommand)
        assert command_table['my-cmd'].name == 'my-cmd'

    def test_main_ops_skips_covered_but_initializes_bare_prefix(
        self, mock_module
    ):
        # Entries registered against bare 'building-command-table' must
        # still be initialized even when main_ops are applied.
        registry = {
            'building-command-table': [
                ('global.module', 'register_global'),
            ],
            'building-command-table.main': [
                ('heavy.module', 'register_add_cmd'),
            ],
        }
        main_ops = [
            (CommandTableOp.ADD, 'my-cmd', 'heavy.module', 'MyCommand'),
        ]
        emitter = LazyInitEmitter(main_command_table_ops=main_ops)
        emitter.load_registry(registry)

        command_table = {}
        session = MagicMock()

        mock_global = MagicMock()
        mock_global.register_global = MagicMock()

        def import_side_effect(mod_path):
            # Ensures that no module besides global.module
            # (including heavy.module) are imported. heavy.module should not be
            # imported because it is present in main_command_table_ops.
            if mod_path == 'global.module':
                return mock_global
            raise AssertionError(f'Unexpected import of {mod_path!r}')

        with patch('importlib.import_module', side_effect=import_side_effect):
            emitter.emit(
                'building-command-table.main',
                command_table=command_table,
                session=session,
            )

        mock_global.register_global.assert_called_once()
        assert emitter.initialized_count == 2
